# -*- coding: utf-8 -*-
"""
Created on Tue Jun 26 10:51:35 2018

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""


def is_int(val):
    """Test whether val is integer
    """
    try : 
        val = float(val)
    except:
        return False
    else:        
        r = int(val) - float(val)
        if r == 0 : return True
        else: return False

def try_int(val):
    """Return int(val) if val is integer, returns val otherwise.
    """
    if is_int(val):
        return int(float(val))
    else: return val

def is_num(val):
    """Return float(val) if val is numeric, returns val otherwise.
    """
    try : val = float(val)
    except : 
        return False
    return True

def try_num(val):
    """Return int(val)/float(val) if val is integer/float, returns val otherwise.
    """
    if is_int(val):
        return int(float(val))
    elif is_num(val):
        return float(val)
    else: return val
    
def remove_list(list_of_list, empty=None, keep_index=0):
    """Remove lists contained in another list.
    """    
    if isinstance(list_of_list, str) :
        list_of_list = str_to_list(list_of_list)
    # if we have a lists from inside a list
    if hasattr(list_of_list,'__iter__') and not isinstance(list_of_list, str):
        new_list = []
        for li in list_of_list:
            if isinstance(li, str):
                li = str_to_list(li)
            while hasattr(li,'__iter__') and not isinstance(li,str):
                if len(li) > 0 : li = li[keep_index]
                else: li = empty
            # check for empty strings
            if isinstance(li,str) and not li:
                li = empty
            new_list.append(li)                
        if len(new_list) == 0:
            new_list = empty
        elif len(new_list) == 1:
            new_list = new_list[0]
        return new_list    
    else: 
        return list_of_list
    
def in_list(anything):
    """Return anything in a list.
    """    
    if hasattr(anything,'__iter__'):
        if not isinstance(anything, str):
            return list(anything)
        elif len(anything) == 0: 
            return []        
        else: 
            return [anything]
    else: 
        return [anything]

def in_array(anything):
    """Return anything in a numpy array.
    """    
    import numpy as np
    if hasattr(anything,'__iter__'):
        if not isinstance(anything, str):
            return np.array(anything)
        elif len(anything) == 0: 
            return np.array([])        
        else: 
            return np.array(anything)
    else: 
        return np.array(anything)

def get_int_list(code, ignore_starting=[], ignore_space=True):
    """Return a list with only integers, extracted from list of strings.
        e.g., ['S 1', 'S 2', 'S 1'] returns [1,2,1].
    """    
    int_idx = []
    int_val = []
    for idx,c in enumerate(code):
        if ignore_space is True:
            c = c.replace(' ','')
        for i in in_list(ignore_starting):
            if c.find(i) == 0:
                c = c.replace(i,'')
        if is_int(c):
            int_idx.append(idx)
            int_val.append(int(c))
    return (int_idx,int_val)

def str_to_list(any_string, opener='[', closer=']'):
    """Return a list extracting from a string.
    """    
    if (len(any_string) > 0) and (any_string[0] == opener) and (any_string[-1] == closer):
        return [try_num(_) for _ in any_string[1:-1].split(',')]
    else:
        return(any_string)

#def getIntValues(code, ignore_starting=[], ignore_space=True):
#    int_idx = getIntIdx(code, ignore_starting=ignore_starting, ignore_space=ignore_space)
#    for idx in int_idx:
#        if ignore_space is True:
#            c = c.replace(' ','')
#        for i in utils.inList(ignore_starting):
#            if c.find(i) == 0:
#                c = c.replace(i,'')
#                break
        


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 12 11:27:20 2018

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""

from .. import emgtools
from . import utilsfunc

def load_txt_file(fname, sep='\t', ch_names_line=None, headerlines=1):
    """Load signal from text file.

    Load signal in text file, organised with channels in columns and time sample
    in rows.

    Parameters
    ----------
    fname : str
        Name of the file to load.
    sep : str
        Separator between columns (default "\t").
    ch_names_line : int
        Line in text file containing channel names. If None, channel names are
        ['channel0', 'channel1', ...] (default None).
    headerlines : int
        Number of headerlines to skip (default 1).

    Returns
    -------
    array : 2D array
        Channels on axis 0, time samples in axis 1.
    ch_dict : dict
        Channels dictionnaries, with channel names as keys and channel index as 
        values.
    """
    import numpy as np
    
    f = open(fname, 'r')
    data_txt = f.read()
    f.close()

    data_list = data_txt.split('\n')
    data_list.remove('')
    
    nr_ts = len(data_list[headerlines:])
    nr_channels = len(data_list[headerlines:][0].split(sep))
    array = np.empty((nr_channels, nr_ts))

    for r in range(nr_ts):
        data_line = data_list[headerlines + r].split(sep)
        for c in range(nr_channels):
            array[c, r] = data_line[c]
        
    ch_dict = {}    
    if ch_names_line is None:
        list_chan = ['channel' + str(c) for c in range(nr_channels)]
    else:
        list_chan = data_list[ch_names_line].split(sep)

    for c in range(nr_channels):
        ch_dict[list_chan[c]] = c 
        
    return array, ch_dict

def get_ch_idx(ch_dict, ch_names='all'):
    """Get channel indices.
    """
    if ch_names == 'all': 
        ch_idx = list(ch_dict.values())
    else: 
        ch_names = utilsfunc.in_list(ch_names)
        ch_idx = [ch_dict[ch] for ch in ch_names if ch in ch_dict.keys()]
    return ch_idx

#def getTimesIdx(struct, tmin = 'first', tmax = 'last'):
#
#    if tmin == 'first' : tmin = 0
#    if tmax == 'last' : tmax = struct.times[-1]
#    times_idx = np.arange(struct.time_as_index(tmin),struct.time_as_index(tmax))
#    return times_idx

def apply_filter(data, ch_dict, sf, ch_names='all', N=3, low_cutoff=None, high_cutoff=None):
    """Apply high and low pass filters to input signal.
    
    Parameters
    ----------
    data : 2D array
        Array containing input signal to filter, channels are on axis 0, 
        time samples on axis 1.
    ch_dict: dict
        Channel names / channel index correspondance dictionnary.
    sf : int | float
        Input signal sampling frequency.
    ch_names : list  | str
        List of channel names on which filter will be applied. If 'all',
        filter is applied on all channels (default 'all').
    N : int
        The order of the filter (default 3).
    low_cutoff : float
        Cutoff frequency for high pass filter. If 'None', no high pass filter
        is applied.         
    high_cutoff : float
        Cutoff frequency for low pass filter. If 'None', no low pass filter
        is applied.
        
    Returns
    -------
    data_filter : 2D array 
        Data array containing filtered signal.
    """
    ch_idx = get_ch_idx(ch_dict, ch_names=ch_names)

    for ch in ch_idx:
        if low_cutoff is not None:
            data[ch] = emgtools.hpfilter(data[ch], N=N, sf=sf, cutoff=low_cutoff)
        if high_cutoff is not None:
            data[ch] = emgtools.lpfilter(data[ch], N=N, sf=sf, cutoff=high_cutoff)
    return data
    
def bipolar_ref(data, ch_dict, anode, cathode, new_ch=None):
    """Set bipolar montage.
    
    Parameters
    ----------
    data : 2D array
        Array containing input signal, channels are on axis 0, 
        time samples on axis 1.
    ch_dict: dict
        Channel names / channel index correspondance dictionnary.
    anode : list  | str
        List of channel names for anode.
    cathode : list  | str
        List of channel names for cathode.
    new_ch : list  | str 
        List of the new channel names. If None, return anode names (default None).
        
    Returns
    -------
    date : 2D array
        Channels on axis 0, time samples in axis 1.
    ch_dict : dict
        Channels dictionnaries, with channel names as keys and channel index as 
        values.
    """
    import numpy as np
    ch_dict = ch_dict.copy()
    anode_idx = get_ch_idx(ch_dict, anode)
    cathode_idx = get_ch_idx(ch_dict, cathode)
    if len(cathode_idx) == 1 :
        cathode_idx = cathode_idx * len(anode_idx)
    if new_ch is None :
        new_ch = anode_idx

    if len(anode_idx) != len(cathode_idx):
        raise ValueError('Number of anode and cathode must be equal, or number of cathode must be 1.')
    if len(new_ch) != len(anode_idx):
        raise ValueError('new_ch names must include new name for each new channel.')
    
    for ch in range(len(anode_idx)):
        
        bipolar_signal = data[anode_idx[ch]] - data[cathode_idx[ch]]
        data = np.vstack((data, bipolar_signal))
        ch_dict[new_ch[ch]] = data.shape[0] - 1
        
    return drop_channels(data, ch_dict, anode + cathode)        

def select_channels(data, ch_dict, ch_names):
    """Select designated channels.
    """
    import numpy as np
    ch_idx = get_ch_idx(ch_dict, ch_names)
    select_data = np.empty((len(ch_idx), data.shape[1]))
    select_ch_dict = {}
    for ch in range(len(ch_idx)):
        select_data[ch] = data[ch_idx[ch]]
        select_ch_dict[ch_names[ch]] = ch
    return select_data, select_ch_dict

def drop_channels(data, ch_dict, ch_names):
    """Delete designated channels.
    """
    chan_list = list(ch_dict.keys())
    select_ch_names = [ch for ch in chan_list if ch not in ch_names]
    return select_channels(data, ch_dict, select_ch_names)
    



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 19 16:24:35 2018

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""   


from .utils import utilsfunc
from . import latency 

def load_continuous(fname, sep=',', sf=None, col_sample=None, col_time=None, col_code=None, \
                    col_chan=None, headerlines=1, bottomlines=0):
    """Load continuous events.

    Load continuous events from text file fname. Fname must contain at least:
    - one column containing events samples or events times
    - one column containing events code
    - one column containing events channel

    Additional columns in fname can contain the events' trial index.

    Parameters
    -----------
    fname : str
        Name of the file to load.
    sep : str
        Separator between columns (default ",").
    sf : int
        Sampling frequency (default None).
    col_sample : int
        The column number containing events time sample (default None).
    col_time : int
        The column number containing events time (default None).
    col_code : int
        The column number containing events code (default None).
    col_chan : int
        The column number containing events channel name/number (default None).
    headerlines:
        Number of headerlines to skip (default 1).
    bottomlines : int
        Number of bottomlines to skip (default 0).

    Returns
    -------
    class Events
        Events object containing read events.
    """
    f = open(fname, 'r')
    evt_txt = f.read()
    f.close()

    evt_list = evt_txt.split('\n')[headerlines:]
    evt_list = evt_list[:len(evt_list)-bottomlines]
    while '' in evt_list:
        evt_list.remove('')

    list_trial = [None, None, None, None, None]
    idx_sample = 0
    idx_time = 1
    idx_code = 2
    idx_chan = 3
    if col_sample is not None:
        list_trial[idx_sample] = []
    if col_time is not None:
        list_trial[idx_time] = []
    if col_code is not None:
        list_trial[idx_code] = []
    if col_chan is not None:
        list_trial[idx_chan] = []

    r = 0
    while r < len(evt_list):

        line = evt_list[r].split(sep)

        if col_sample is not None:
            list_trial[idx_sample].append(utilsfunc.try_int(line[col_sample]))
        if col_time is not None:
            list_trial[idx_time].append(float(line[col_time]))
        if col_code is not None:
            list_trial[idx_code].append(utilsfunc.try_int(line[col_code]))
        if col_chan is not None:
            list_trial[idx_chan].append(utilsfunc.try_int(line[col_chan]))

        r += 1

    return Events(sample=list_trial[idx_sample], time=list_trial[idx_time],\
                  code=list_trial[idx_code], chan=list_trial[idx_chan], sf=sf)


def load_segmented(fname, headerlines=1, sep=',', sf=None,\
                   col_sample=None, col_time=None, col_code=None, col_chan=None,\
                   col_trial_idx=None, \
                   col_tmin_sample=None, col_tmin_time=None,\
                   col_t0_sample=None, col_t0_time=None, \
                   col_tmax_sample=None, col_tmax_time=None):
    """Load segmented events.

    Load segmented events from text file fname. Fname must contain at least:
    - one column containing events samples or events times
    - one column containing events code
    - one column containing events channel
    - one column containing trial index

    Additional columns can contain the time of reference of each trial in the
    corresponding continuous raw file:
    - the sample corresponding to trial's start time in continuous (tmin sample)
    - the time corresponding to trial's start time in continuous (tmin time)
    - the sample corresponding to trial's time 0 in continuous (t0 sample)
    - the time corresponding to trial's time 0 in continuous (t0 time)
    - the sample corresponding to trial's end time in continuous (tmax sample)
    - the time corresponding to trial's end time in continuous (tmax time)

    It is necessary to provide sampling frequency.

    Parameters
    -----------
    fname : str
        Name of the file to load.
    headerLines : int
        Number of headerlines to skip (default 1).
    sep : str
        Separator between columns (default ",").
    sf : int
        Sampling frequency (default None).
    col_code : int
        The column number containing events code (default None).
    col_sample : int
        The column number containing events time sample (default None).
    col_sample : int
        The column number containing events time (default None).
    col_chan : int
        The column number containing events channel name/number (default None).
    col_trial_idx : int
        The column number containing trial index (default None).
    col_tmin_sample: int
        The column number containing sample of trial's start time (default None).
    col_tmin_time : int
        The column number containing time of trial's start time (default None).
    col_t0_sample : int
        The column number containing sample of trial's time 0 (default None).
    col_t0_time : int
        The column number containing time of trial's time 0 (default None).
    col_tmax_sample: int
        The column number containing sample of trial's end time (default None).
    col_tmax_time : int
        The column number containing time of trial's end time (default None).

    Returns
    -------
    class EpochEvents
        EpochEvents object containing read events.
    """
    f = open(fname, 'r')
    evt_txt = f.read()
    f.close()

    evt_list = evt_txt.split('\n')[headerlines:]
    evt_list.remove('')

    try:
        col_trial_idx = int(col_trial_idx)
    except:
        print('Trial index column is necessary, provide valid integer value for col_trialIdx.')
        raise

    ep_evts = EpochEvents(sf=sf)

    list_trial = [None, None, None, None]
    idx_sample = 0
    idx_time = 1
    idx_code = 2
    idx_chan = 3
    if col_sample is not None:
        list_trial[idx_sample] = []
    if col_time is not None:
        list_trial[idx_time] = []
    if col_code is not None:
        list_trial[idx_code] = []
    if col_chan is not None:
        list_trial[idx_chan] = []

    # we read once to determine code and channel types (e.g., int, str ...)
    r = 0
    type_code = 'numeric'
    type_chan = 'numeric'
    while (r < len(evt_list)) and ((type_code ==  'numeric') or (type_chan == 'numeric')):
        line = evt_list[r].split(sep)
        if (type_code == 'numeric') and not (utilsfunc.is_num(line[col_code])):
            type_code = 'string'
        if (type_chan == 'numeric') and not (utilsfunc.is_num(line[col_chan])):
            type_chan = 'string'
        r+=1

    sample_latencies = [None, None, None]
    time_latencies = [None, None, None]

    r = 0
    while r < len(evt_list):
        
        line = evt_list[r].split(sep)
        t = int(line[col_trial_idx])

        if col_sample is not None:
            list_trial[idx_sample] = [int(line[col_sample])]
        if col_time is not None:
            list_trial[idx_time] = [float(line[col_time])]
        if col_code is not None:
            if type_code == 'numeric':
                list_trial[idx_code] = [utilsfunc.try_num(line[col_code])]
            else:
                list_trial[idx_code] = [line[col_code]]
        if col_chan is not None:
            if type_chan == 'numeric':
                list_trial[idx_chan] = [utilsfunc.try_num(line[col_chan])]
            else:
                list_trial[idx_chan] = [line[col_chan]]
                
        if col_tmin_sample is not None:
            sample_latencies[0] = int(line[col_tmin_sample])
        if col_tmin_time is not None:
            time_latencies[0] = float(line[col_tmin_time])
        if col_t0_sample is not None:
            sample_latencies[1] = int(line[col_t0_sample])
        if col_t0_time is not None:
            time_latencies[1] = float(line[col_t0_time])
        if col_tmax_sample is not None:
            sample_latencies[2] = int(line[col_tmax_sample])
        if col_tmax_time is not None:
            time_latencies[2] = float(line[col_tmax_time])

        tmin = latency.Latency(sample=sample_latencies[0], time=time_latencies[0], sf=ep_evts.sf)
        t0 = latency.Latency(sample=sample_latencies[1], time=time_latencies[1], sf=ep_evts.sf)
        tmax = latency.Latency(sample=sample_latencies[2], time=time_latencies[2], sf=ep_evts.sf)
        
        r += 1
        while (r < len(evt_list)) and (t == int(evt_list[r].split(sep)[col_trial_idx])): 
            line = evt_list[r].split(sep)
            t = int(line[col_trial_idx])
            if col_sample is not None:
                # sample must be an integer
                list_trial[idx_sample].append(int(line[col_sample]))
            if col_time is not None:
                # time must be a float
                list_trial[idx_time].append(float(line[col_time]))
            if col_code is not None:
                # code can be numeric/integer or string
                if type_code == 'numeric':
                    list_trial[idx_code].append(utilsfunc.try_num(line[col_code]))
                else:
                    list_trial[idx_code].append(line[col_code])
            if col_chan is not None:
                # chan can be numeric/integer or string
                if type_chan == 'numeric':
                    list_trial[idx_chan].append(utilsfunc.try_num(line[col_chan]))
                else:
                    list_trial[idx_chan].append(line[col_chan])
            r += 1
                           
        evt_trial = Events(sample=list_trial[idx_sample], time=list_trial[idx_time], \
                           code=list_trial[idx_code], chan=list_trial[idx_chan], \
                           sf=ep_evts.sf)
        ep_evts.add_trial(evt_trial, latency_min=tmin, latency0=t0,\
                          latency_max=tmax)

    return ep_evts


##############################################################################
# Class Events and EpochEvents ###############################################
##############################################################################
    
class Events():
    """Events class, storing and manipulating event markers.

    'Events' contains the latency (time and sample), the code and the channel
    of event markers (typically all events of one experiment). Information is
    stored in numpy.ndarrays (Events.lat.sample, Events.lat.time,
    Events.code and Events.chan), in which one given event is stored at the
    same position in all arrays (i.e., information for event 'n' is stored in
    Events.lat.sample[n], Events.lat.sample[n], Events.code[n] and
    Events.chan[n]). For Latency, either time (in seconds) or sample can be
    given. Time/sample, code and chan must be of same length. If sampling frequency
    (sf) is not given, it will be asked when the object is created.

    Parameters
    ----------
    sample : int | list of int | 1D array of int
        Latency in samples of each event marker.
    time : float | list of float | 1D array of float
        Latency in second of each event marker.
    code : int | float | str | list | 1D array
        Code of each event marker.
    chan : int | str | list | 1D array
        Channel of each event marker.
    sf : float
        Sampling frequency of the corresponding data recording.
    """

    def __init__(self, sample=None, time=None, code=None, chan=None, sf=None):
        import numpy as np

        if sf is None:
            sf = input('Enter sampling frequency: ')
        self.sf = float(sf)
        self.lat = latency.Latency(sample=sample, time=time, sf=self.sf)

        if chan is None:
            self.chan = np.array([], dtype=int)
        else: self.chan = np.array(utilsfunc.in_list(chan))

        if code is None:
            self.code = np.array([], dtype=int)
        else: self.code = np.array(utilsfunc.in_list(code))

        if ((self.lat.sample.size) != (self.chan.size)) | \
           ((self.lat.sample.size) != (self.code.size)):
            raise TypeError('sample/time, chan and code must be of same length.')

    def __repr__(self):
        return "class Events, {} events, sf = {} \n latency: {} \n code: {} \n chan: {}".format(self.nb_events(), self.sf, self.lat, self.code, self.chan)

    def get_events(self, idx_evt):
        """Get the events designated by idxEvt and returns them.

        Parameters
        ----------
        idx_evt : int | list of int
            Position(s) of the event(s) to get.

        Returns
        -------
            Events containing only the designated events.
        """
        return Events(sample=self.lat.sample[idx_evt], time=self.lat.time[idx_evt], code=self.code[idx_evt], chan=self.chan[idx_evt], sf=self.sf)

    def del_events(self, idx_evt, print_del_evt=True):
        """Delete the events designated by idxEvt, occurs in place.

        Parameters
        ----------
        idx_evt : int | list of int
            Position(s) of the event(s) to delete.
        print_del_evt : bool
            If True, print the number of deleted events.
        """

        import numpy as np
        idx_evt = utilsfunc.in_list(idx_evt)
        to_print = str(len(idx_evt)) + ' event(s) removed.'
        if print_del_evt is True:
            print(to_print)

        self.lat.del_lat(idx_evt)
        self.code = np.delete(self.code, idx_evt)
        self.chan = np.delete(self.chan, idx_evt)

    def nb_events(self):
        """Return the number of single events.
        """
        return len(self.code)
 
    def sort_events(self):
        """Sort the single events by latency, occurs in place.
        """
        sort_sample = self.lat.sort_sample()
        self.lat.sample = self.lat.sample[sort_sample]
        self.lat.time = self.lat.time[sort_sample]
        self.code = self.code[sort_sample]
        self.chan = self.chan[sort_sample]
        
    def find_events(self, evt=None, sample=None, time=None, code=None, chan=None,\
                    tmin=None, tmax=None, samplemin=None, samplemax=None):
        """Find the specified event(s) and return the position(s).

        Finds the event(s) whose sample, time, chan or code is equal to the
        given value(s) (e.g., Events.find_events(code='onset') will return
        positions of all events whose code is 'onset'). If more than one
        attribute is filled, finds events satisfying both requirements
        (Events.find_events(code='onset', chan=1) will return positions of
        events whose code is 'onset' AND chan is 1). If more than one value
        is given for one attribute, finds events whose attribute is equal to
        either value (Events.find_events(code='onset', chan=[0,1]) will return
        positions of events whose code is 'onset' and chan is 0 OR 1).
        See also Latency.find_lat for sample and time parameters.
        To find event(s) in a restricted time window, use tmin or samplemin
        and tmax or samplemax parameters.

        Parameters
        ----------
        evt : class Events
            If evt is given, finds events with the same Latency, code and chan.
            In this case, other parameters are ignored.
        sample : int | list of  int | 1D array of int
            Sample value(s) to look for.
        time : float | list of float | 1D array of float
            Time value(s) to look for.
        code : int | str | list | 1D array
            Code value(s) to look for.
        chan : int | str | list | 1D array
            Channel value(s) to look for.
        tmin : float
            Only events occurring after time latency tmin are returned.
        tmax : float
            Only events occurring before time latency tmax are returned.
        samplemin : int
            Only events occurring after sample latency samplemin are returned.
        samplemax : int
            Only events occurring before sample latency samplemax are returned.

        Returns
        -------
            Position(s) of the found events.
        """

        import numpy as np
        
        if evt is not None:
            
            if (sample is not None) | (time is not None)\
                | (chan is not None) | (code is not None):
                    raise ValueError ('Either evt OR sample/time/code/chan can be searched.')

#            # first selection on latencies events to corresponding latencies
            idx_select = []
            for e in range(evt.nb_events()):
                sample_index = np.flatnonzero(self.lat.sample == evt.lat.sample[e])
                # keeps only those whose time, code and chan are also equal
                idx_select.extend([s for s in sample_index if (np.abs(self.lat.time[s] - evt.lat.time[e]) < 10**(-10))  \
                                                             & (self.code[s] == evt.code[e]) \
                                                             & (self.chan[s] == evt.chan[e])])
#            idx_select = np.sort(np.asarray(idx_select, dtype=np.int64))

        elif (sample is None) & (time is None) & (code is None) & (chan is None)\
            & (tmin is None) & (tmax is None)\
            & (samplemin is None) & (samplemax is None):
            idx_select = np.flatnonzero(np.zeros(1))
        else:
            select = np.ones(self.nb_events(), dtype=bool)
            if sample is not None:
                select = select & np.isin(self.lat.sample, sample)
            if time is not None:
                select = select & np.isin(self.lat.time, time)

            if tmin is not None:
                select = select & (self.lat.time >= tmin)
            if tmax is not None:
                select = select & (self.lat.time <= tmax)
            if samplemin is not None:
                select = select & (self.lat.sample >= samplemin)
            if samplemax is not None:
                select = select & (self.lat.sample <= samplemax)

            if code is not None:
                select = select & np.isin(self.code, code)
            if chan is not None:
                select = select & np.isin(self.chan, chan)
                
                
            idx_select = np.flatnonzero(select)
            
        # if tmin is not None:
        #     idx_select = idx_select[self.lat.time[idx_select]>=tmin]
        # if tmax is not None:
        #     idx_select = idx_select[self.lat.time[idx_select]<=tmax]

        return np.sort(np.asarray(idx_select, dtype=np.int64))    
    
    
    def add_events(self, evt, sort=True, drop_duplic=False):
        """Add events, occurs in place.

        Adds events to the Events object. If sort is True, events are then
        sorted by latency. If drop_duplic is True, removes duplicated after
        adding events.

        Parameters
        ----------
        evt : class Events
            Contains the event(s) to add.
        sort : bool
            If True, events are sorted by latency after adding the new events.
        drop_duplic : bool
            If True, new events with same time/sample, code and chan are removed
            using drop_duplicates().
        """
        import numpy as np
        self.lat.add_lat(evt.lat)
        self.code = np.concatenate((self.code, evt.code))
        self.chan = np.concatenate((self.chan, evt.chan))
        
        if drop_duplic:
            self.drop_duplicates()
        if sort:
            self.sort_events()

    def find_and_get_events(self, evt=None, sample=None, time=None, code=None, chan=None,\
                            tmin=None, tmax=None, samplemin=None, samplemax=None, print_find_evt=True):
        """Return Events containing only the specified events.

        Use find_events to find designated events and return new Events,
        containing only the selected events.

        Parameters
        ----------
        evt : class obj Events
            If evt is given, selects events with the same Latency, code and chan.
        sample : int | list of int | 1D array of int
            Sample value(s) to select.
        time : float | list of float | 1D array of float
            Time value(s) to select.
        code : int | str | list | 1D array
            Code value(s) to select.
        chan : int | str | list | 1D array
            Channel value(s) to select.
        tmin : float
            Only events occurring after time latency tmin are returned.
        tmax : float
            Only events occurring before time latency tmax are returned.
        samplemin : int
            Only events occurring after sample latency samplemin are returned.
        samplemax : int
            Only events occurring before sample latency samplemax are returned.
        print_find_evt : bool
            If True, print the number of selected events.

        Returns
        -------
            Events containing only the designated events.
        """
        idx_evt = self.find_events(evt=evt, sample=sample, time=time, chan=chan, code=code,\
                                   tmin=tmin, tmax=tmax, samplemin=samplemin, samplemax=samplemax)

        if len(idx_evt) > 0:
            to_print = str(len(idx_evt)) + ' event(s) selected.'
        else:
            to_print = 'Event(s) not found, nothing was selected.'
        if print_find_evt is True:
            print(to_print)
        return self.get_events(idx_evt)
            
    def find_and_del_events(self, evt=None, sample=None, time=None, code=None, chan=None,\
                            tmin=None, tmax=None, samplemin=None, samplemax=None, print_del_evt=True):
        """Remove the specified events, occur in place.

        Use find_events to find designated events and removes them from the
        current Events object.

        Parameters
        ----------
        evt : class Events
            If evt is given, removes events with the same Latency, code and
            chan.
        sample : int | list of int | 1D array of int
            Sample value(s) of events to remove.
        time : float | list of float | 1D array of float
            Time value(s) of events to remove.
        code : int | str | list of int | list of str
            Code value(s) of events to remove.
        chan : int | str | list of int | list of str
            Channel value(s) of events to remove.
        tmin : float
            Only events occurring after time latency tmin are returned.
        tmax : float
            Only events occurring before time latency tmax are returned.
        samplemin : int
            Only events occurring after sample latency samplemin are returned.
        samplemax : int
            Only events occurring before sample latency samplemax are returned.
        print_del_evt : bool
            If True, print the number of deleted events.
        """
        idx_evt = self.find_events(evt=evt, sample=sample, time=time, code=code, chan=chan,\
                                   tmin=tmin, tmax=tmax, samplemin=samplemin, samplemax=samplemax)
        self.del_events(idx_evt, print_del_evt=print_del_evt)

    def copy(self):
        """Return a copy of the current Events.
        """
        import numpy as np
        copy_sample = np.copy(self.lat.sample)
        copy_time = np.copy(self.lat.time)
        copy_code = np.copy(self.code)
        copy_chan = np.copy(self.chan)
        return Events(sample=copy_sample, time=copy_time, code=copy_code, chan=copy_chan, sf=self.sf)

    def to_csv(self, fname, header="sample,time,code,chan", sep=',', save_sample=True, save_time=True, save_code=True, save_chan=True, save_trial_idx=False, trial_idx=None):
        """Export events in fname, csv format.

        Parameters
        ----------
        fname : str
            Name of the file to save.
        header : str
            String for the first line in csv file, should contain column names
            (default "sample,time,code,chan").
        sep : str
            Separator between columns (default ",").
        save_sample : bool
            If True, save the column containing event sample (default True).
        save_time : bool
            If True, save the column containing event time (default True).
        save_code : bool
            If True, save the column containing event code (default True).
        save_chan : bool
            If True, save the column containing event channel (default True).
        save_trial_idx : bool
            If True, save a column containing event trial index. Can be True 
            only for events which have been segmented already (default False).
        trial_idx : list | 1D array
            Contain the trial index for each event, should be provided
            when save_trial_idx is set to True.
        """
        import numpy as np

        idx_col = np.where([save_sample, save_time, save_code, save_chan, save_trial_idx])[0]
        list_col = [self.lat.sample, self.lat.time, self.code, self.chan, trial_idx]

        f = open(fname, 'w')
        f.write(header + '\n')
        for e in range(self.nb_events()):
            evt = [str(list_col[c][e]) for c in idx_col]
            for c in evt[:-1]:
                f.write(c + sep)
            f.write(evt[-1] + '\n')
        f.close()       
        
    def to_txt(self, fname, event_type={}, chan_names={}, header="Type, Description, Position, Length, Channel", sep=','):
        """Export events in fname, text format to be read in BVA.

        Parameters
        ----------
        fname : str
            Name of the file to save.
        event_type : dict
            Specify the type of marker (e.g., 'Stim', 'EMG') for each marker
            code. Type 'Comments' will be assigned to markers with no type specified.
        header : str
            String for the first line in csv file, should contain column names
            (default "Type, Description, Position, Length, Channel").
        sep : str
            Separator between columns (default ",").
        """
        length = [1] * self.nb_events()

        f = open(fname, 'w')
        f.write('Sampling rate: ' + str(int(self.sf)) + 'Hz, SamplingInterval: ' + str(1/self.sf*1000) + 'ms\n')
        f.write(header + '\n')

        for e in range(self.nb_events()):
            if self.code[e] in event_type.keys():
                e_type = event_type[self.code[e]]
            else:
                e_type = 'Comments'
            if self.chan[e] in chan_names.keys():
                chan = chan_names[self.chan[e]]
            else:
                chan = self.chan[e]

            try:
                int(self.code[e])
            except ValueError:
                code = str(self.code[e])
            else:
                code = str(int(self.code[e]))

            f.write(e_type + sep + code + sep + str(int(self.lat.sample[e]))\
                    + sep + str(length[e]) + sep + str(chan) +'\n')
        f.close()

    def to_bva_vmrk(self, fname, event_type={}, chan_names={}):
        """Save file in vmrk format from Brain Vision (format '.vmrk').

        Parameters
        ----------
        fname : str
            Name of the vmrk file to save.
        event_type : dict
            Specify the type of marker (e.g., 'Stim', 'EMG') for each marker
            code. Type 'Comments' will be assigned to markers with no type specified.
        chan_names : dict
            Specify the name of the channel for each channel number in events
            (e.g., chan_names = {-1: 'All'}). Channel numbers are used when no
            channel name is specified.
        """
        import os
        core_name = os.path.split(fname)[-1]
        vmrk_name = fname + '.vmrk'
    #    with codecs.open(vhdrFilename, "w", encoding="utf-8") as f:
    #        muv = u'µ'

        f = open(vmrk_name, 'w')
        f.write('Brain Vision Data Exchange Marker File, Version 2.0\n')
        f.write('; Create from to_bva_vmrk in events.py\n')
        f.write('; The channel numbers are related to the channels in the exported file.')
        f.write('\n')

        f.write('[Common Infos]\n')
        f.write('Codepage=UTF-8\n')
        f.write('DataFile='+core_name+'.eeg\n')
        f.write('\n')

        f.write('[Marker Infos]\n')
        for e in range(self.nb_events()):
            if self.code[e] in event_type.keys():
                e_type = event_type[self.code[e]]
            else:
                e_type = 'Comments'
            if self.chan[e] in chan_names.keys():
                chan = chan_names[self.chan[e]]
            else:
                chan = self.chan[e]

            try:
                int(self.code[e])
            except ValueError:
                code = str(self.code[e])
            else:
                code = str(int(self.code[e]))

            f.write('Mk'+str(e+1)+'='+ e_type +','+code+\
            ','+ str(int(self.lat.sample[e]))+',1,' + str(chan) +'\n')
        f.write('\n')

        f.write('[Marker User Infos]\n')
        f.write('\n')

        f.close()

    def to_bva_markers(self, fname, event_type={}, chan_names={}):
        """Save file in vmrk format from Brain Vision (format '.Markers').

        Parameters
        ----------
        fname : str
            Name of the vmrk file to save.
        event_type : dict
            Specify the type of marker (e.g., 'Stim', 'EMG') for each marker
            code. Type 'Comments' will be assigned to markers with no type specified.
        chan_names : dict
            Specify the name of the channel for each channel number in events
            (e.g., chan_names = {-1: 'All'}). Channel numbers are used when no
            channel name is specified.
        """
        markers_name = fname + '.Markers'

        f = open(markers_name, 'w')
        f.write(('Sampling rate: {}Hz, SamplingInterval: {}ms\n').format(self.sf,1000/self.sf))
        f.write('Type, Description, Position, Length, Channel\n')

        for e in range(self.nb_events()):
            if self.code[e] in event_type.keys():
                e_type = event_type[self.code[e]]
            else:
                e_type = 'Comments'
            if self.chan[e] in chan_names.keys():
                chan = chan_names[self.chan[e]]
            else:
                chan = self.chan[e]

            try:
                int(self.code[e])
            except ValueError:
                code = str(self.code[e])
            else:
                code = str(int(self.code[e]))

#            f.write(eType+', ' + str(e+1)+'='+ eType +','+code+\
#            ','+ str(int(self.lat.sample[e]))+',1,' + str(chan) +'\n')
            f.write(('{}, {}, {}, 1, {}\n').format(e_type,
                                                   code,
                                                   self.lat.sample[e],
                                                   chan))
        f.write('\n')
        f.close()
        
    def get_duplicates(self):
        """Get duplicated events (i.e., same sample/time, code and chan).
        """
        import numpy as np
        print("Checking for duplicates in events...")
        first_evt_idx = np.array([], dtype=int)
        duplicated_evt_idx = np.array([], dtype=int)
        for e in range(self.nb_events()):
            if not (e == duplicated_evt_idx).any():
                evt = self.get_events(e)
                evt_idx = self.find_events(evt=evt)
                if len(evt_idx) > 1:
                    first_evt_idx = np.concatenate((first_evt_idx,[e]*(len(evt_idx)-1)))
                    duplicated_evt_idx = np.concatenate((duplicated_evt_idx,evt_idx[evt_idx!=e]))
                    
        return np.vstack((first_evt_idx,duplicated_evt_idx)).T

    def drop_duplicates(self):
        """Delete duplicated events (i.e., same sample/time, code and chan).
        """
        duplicates = self.get_duplicates()
        self.del_events(duplicates[:,1])            

        
    def segment(self, code_t0=None, pos_events0=None, tmin=0, tmax=1, print_epochs=True):
        """Segment events on the interval [tmin,tmax] around code_t0 events.

        Epoch segmentation: events are segmented on the interval between tmin
        and tmax, around the time latency of the reference (t0) events.
        Reference events are specified either by event code, i.e., events
        whose code is equal to code_t0, or by events positions, i.e., events
        at positions corresponding to pos_events0.
        Events of each segmented epoch are stored in an EpochEvents object
        (class EpochEvents). 

        Parameters
        ----------
        code_t0 : int | str | list
            Code of events to use as reference (t0) events. Does not accept 
            dict_values. 
        pos_events0: int | list
            Position of events to use as reference (t0) events.
        tmin : float
            Start time (in seconds) of the epoch segmentation,
            relative to the reference events (default 0).
        tmax : float
            End time (in seconds) of the epoch segmentation,
            relative to the reference events (default 1s).
        print_epochs : bool
            If True, print the number of epochs (default is True).
        
        Returns
        -------
        ep_evts
            EpochEvents object containing segmented events.
        """
        import numpy as np
        
#        timeSerie = times(tmin, tmax, self.sf)
#        samplemin = timeSerie[0] * self.sf
#        samplemax = timeSerie[-1] * self.sf

        sample_serie = latency.samples(tmin, tmax, self.sf)
        samplemin = sample_serie[0]
        samplemax = sample_serie[-1]

        if pos_events0 is None:
            pos_events0 = []
        else:
            pos_events0 = utilsfunc.in_list(pos_events0)

        if code_t0 is not None:
            pos_events0.extend(self.find_events(code=code_t0))
            
        sample_epochs = self.lat.sample[pos_events0]
        time_epochs = self.lat.time[pos_events0]
#        sampleEpochs.sort()
#        timeEpochs.sort()
        nb_epochs = len(pos_events0)
        if print_epochs is True:
            print('Found {} epoch(s).'.format(nb_epochs))
        
        ep_evts = EpochEvents(sf=self.sf)

        for e in np.arange(nb_epochs):
            idx = np.where((self.lat.sample >= sample_epochs[e] + samplemin) \
                         & (self.lat.sample <= sample_epochs[e] + samplemax))

            evts_trial = Events(sample=self.lat.sample[idx] - (sample_epochs[e] + samplemin), \
                               time=self.lat.time[idx] - time_epochs[e],\
                               chan=self.chan[idx], code=self.code[idx], sf=self.sf)
            ep_evts.add_trial(evts_trial, latency_min=latency.Latency(sample=sample_epochs[e] + samplemin,sf=self.sf),\
                              latency0=latency.Latency(sample=sample_epochs[e], sf=self.sf),\
                              latency_max=latency.Latency(sample=sample_epochs[e] + samplemax, sf=self.sf))

        return ep_evts
    

class EpochEvents():
    """EpochEvents class, storing and manipulating epoch event markers.

    Attributes
    ----------
    list_evts_trials : list
        List of Events objects of consecutive trials (i.e., list_evts_trials[n]
        contains trial n). Latency in single trials' Events is relative to the
        trial's time zero (t0) event.
    tmin : class Latency
        Latency object containing starting time and sample of consecutive trials
        (tmin[n] contains starting latency of trial n).Latencies are given
        relative to the original continous Events object.
    t0 : class Latency
        Latency object containing time and sample of the 'time0' event (i.e.,
        the event used for segmentation) in consecutive trials, relative to the
        original continous Events object.
    tmax : class Latency
        Latency object containing ending time and sample of consecutive trials,
        relative to the original continous Events object.

    Parameters
    ----------
    sf : float
        Sampling frequency of the corresponding data recording.
    """

    def __init__(self, sf=None):

        self.list_evts_trials = []
        if sf is None:
            sf = input('Enter sampling frequency: ')
        self.sf = float(sf)
        self.tmin = latency.Latency(sf=self.sf)
        self.t0 = latency.Latency(sf=self.sf)
        self.tmax = latency.Latency(sf=self.sf)

    def __repr__(self):
        return "class EpochEvents, {} trials, {} events, sf = {} ".format(self.nb_trials(), self.nb_events(), self.sf)

    def add_trial(self, evt_trial, latency_min, latency0, latency_max):
        """Add the events trial.
                
        Parameters
        ----------
        evt_trial : class Events 
            Contains the event(s) to add.
        latency_min : class Latency
            Latency object containing starting time and sample of the trial
            to add, relative to the original continous Events object.
        latency0 : class Latency
            Latency object containing time and sample of the 'time0' event 
            of the trial to add, relative to the original continous Events object.
        latency_max : class Latency
            Latency object containing ending time and sample of the trial
            to add, relative to the original continous Events object.
        """
        self.list_evts_trials.append(evt_trial)
        self.tmin.add_lat(latency_min)
        self.t0.add_lat(latency0)
        self.tmax.add_lat(latency_max)


    def del_trial(self, idx_trial):
        """Delete the events trial.
        """
        idx_trial = utilsfunc.in_list(idx_trial)
        for n, trial in enumerate(idx_trial):
            del self.list_evts_trials[trial-n]
        self.tmin.del_lat(idx_trial)
        self.t0.del_lat(idx_trial)
        self.tmax.del_lat(idx_trial)

    def get_trials(self, idx_trial):
        """Return a new EpochEvents with only specified trials.
        """
        new_evts = EpochEvents(sf=self.sf)
        for trial in utilsfunc.in_list(idx_trial):
            new_evts.add_trial(self.list_evts_trials[trial],\
                               latency_min=latency.Latency(sample=self.tmin.sample[trial], sf=self.sf),\
                               latency0=latency.Latency(self.t0.sample[trial], sf=self.sf),\
                               latency_max=latency.Latency(sample=self.tmax.sample[trial], sf=self.sf))
        return new_evts

    def nb_trials(self):
        """Return the number of trials.
        """
        return len(self.list_evts_trials)

    def nb_events(self):
        """Return the number of single Events.
        """
        e = 0
        for trial in range(self.nb_trials()):
            e += self.list_evts_trials[trial].nb_events()
        return e
        
    def as_continuous(self, drop_duplic=True, sort=True):
        """Return continuous Events.

        Extracts single trials' Events and recreates continuous Events, based
        on tmin Latency: sample latencies in the new continuous events are
        computed by adding sample tmin of each trial to the trial events latencies.

        Parameters
        ----------
        drop_duplic : bool
            If True, searchs and removes events duplicates in the new continuous
            Events (e.g., when one trial Events contained some of the next trial
            events) (default True).
        sort : bool
            If True, returns events sorted by latency.
            
        Returns
        -------
        evts_as_continuous : class Events
            Events object containing continuous events.
        trial_idx : 1D array
            Contains the trial index corresponding to each kept single event.
        """
        import numpy as np
        evts_as_continuous = Events(sf=self.sf)
        trial_idx = np.array([], dtype=int)
        for trial in range(self.nb_trials()):
            evts_as_continuous.add_events(Events(sample=self.list_evts_trials[trial].lat.sample + self.tmin.sample[trial],\
                                                 time=self.list_evts_trials[trial].lat.time + self.t0.time[trial],\
                                                 code=self.list_evts_trials[trial].code,\
                                                 chan=self.list_evts_trials[trial].chan, sf=self.sf), sort=False)
            trial_idx = np.append(trial_idx, ([trial] * self.list_evts_trials[trial].nb_events()))

        if drop_duplic:
            
            duplicates = evts_as_continuous.get_duplicates()
            evts_as_continuous.del_events(duplicates[:,1])
            trial_idx = np.delete(trial_idx, duplicates[:,1])

        if sort:
            sorted_idx = evts_as_continuous.lat.sort_sample()
            trial_idx = trial_idx[sorted_idx]
            evts_as_continuous.sort_events()

        return evts_as_continuous, trial_idx

    def continuous_to_csv(self, fname, header="sample,time,code,chan,trialIdx", sep=',', save_sample=True, save_time=True, save_code=True, save_chan=True, save_trial_idx=True):
        """Export continuous events in fname, csv format.

        Parameters
        -----------
        fname : str
            Name of the file to save.
        header : str
            String for the first line in csv file, should contain column names
            (default "sample,time,code,chan").
        sep : str
            Separator between columns (default ",").
        save_sample : bool
            If True, saves the column containing event sample (default True).
        save_time : bool
            If True, saves the column containing event time (default True).
        save_code : bool
            If True, saves the column containing event code (default True).
        save_chan : bool
            If True, saves the column containing event channel (default True).
        save_trial_idx : bool
            If True, saves a column containing event trial index.
        """
        import numpy as np

        print("Saving as continuous...")
        continuous_evts, trial_idx = self.as_continuous()
        idx_col = np.where([save_sample, save_time, save_code, save_chan, save_trial_idx])[0]
        list_col = [continuous_evts.lat.sample, continuous_evts.lat.time, continuous_evts.code, continuous_evts.chan, trial_idx]

        f = open(fname, 'w')
        f.write(header + '\n')
        for e in range(continuous_evts.nb_events()):
            evt = [str(utilsfunc.try_int(list_col[c][e])) for c in idx_col]
            for c in evt[:-1]:
                f.write(c + sep)
            f.write(evt[-1] + '\n')
        f.close()
       
    def to_csv(self, fname, header=["sample","time","code","chan","trial_idx","trial_raw_tmin.sample","trial_raw_tmin.time","trial_raw_t0.sample","trial_raw_t0.time","trial_raw_tmax.sample","trial_raw_tmax.time"],\
               sep=',', save_sample=True, save_time=True, save_code=True, save_chan=True, save_trial_idx=True, save_tmin=True, save_t0=True, save_tmax=True):
        """Export segmented events in fname, csv format.
        
        Parameters
        ----------
        fname : str
            Name of the file to save.
        header : list
            List of head columns to insert as first line in csv file, should 
            contain column names (default ["sample","time","code","chan",
            "trial_idx","trial_raw_tmin.sample","trial_raw_tmin.time",
            "trial_raw_t0.sample","trial_raw_t0.time","trial_raw_tmax.sample",
            "trial_raw_tmax.time"]).
        sep : str
            Separator between columns (default ",").
        save_sample : bool
            If True, saves the column containing event sample (default True).
        save_time : bool
            If True, saves the column containing event time (default True).
        save_code : bool
            If True, saves the column containing event code (default True).
        save_chan : bool
            If True, saves the column containing event channel (default True).
        save_trial_idx : bool
            If True, saves a column containing event trial index (default True).
        save_tmin : bool
            If True, saves the columns containing tmin samples and tmin time
            (default True).
        save_t0 : bool
            If True, saves the columns containing t0 samples and t0 time
            (default True).
        save_tmax : bool
            If True, saves the columns containing tmax samples and tmax time
            (default True).
        """

        import numpy as np

        print("Saving as segmented...")
        idx_evt_col = np.where([save_sample, save_time, save_code, save_chan])[0]
        idx_info_col = np.where([save_trial_idx, save_tmin, save_tmin, save_t0, save_t0, save_tmax, save_tmax])[0]

        f = open(fname, 'w')
        f.write(sep.join(header) + '\n')

        for t in range(self.nb_trials()):
            evt_col = [self.list_evts_trials[t].lat.sample, self.list_evts_trials[t].lat.time, self.list_evts_trials[t].code, self.list_evts_trials[t].chan]
            info_col = [t, self.tmin.sample[t], self.tmin.time[t], self.t0.sample[t], self.t0.time[t], self.tmax.sample[t], self.tmax.time[t]]

            for e in range(self.list_evts_trials[t].nb_events()):
                evt = [str(evt_col[c][e]) for c in idx_evt_col] + [str(info_col[c]) for c in idx_info_col]
                for c in evt[:-1]: f.write(c + sep)
                f.write(evt[-1] + '\n')
        f.close()
    
    def get_data(self, data, drop_trials = []):
        """Get data segments corresponding to EpochEvents.
    
        Parameters
        ----------
        data : 2D array
            Contains continuous data (axis 0: channels, axis 1: time).
        drop_trials : list | 1D array
            Indices of trials that will not be included in data epochs.
    
        Returns
        -------
        epochs: 3D array
            Contains data segments (axis 0: trials, axis 1: channels, axis 2: time).
        """
        import numpy as np
    
        if data.ndim == 1:
            print('Warning, data is 1D array and must be transposed! Ignore me if your data contains only one channel.')
            data = data[np.newaxis]
			
        drop_trials = utilsfunc.in_list(drop_trials)
        trials = np.arange(self.nb_trials())
        for trial in drop_trials:
            if trial not in trials:
                raise IndexError('drop_trials ' + str(trial) + ': out of bound.')
        trials = np.delete(trials, drop_trials)
    
        if (self.tmin.sample[trials] < 0).any():
            raise ValueError('All tmin.sample must be positive, use drop_trials if necessary.')
        elif (self.tmax.sample[trials] > data.shape[1]).any():
            raise ValueError('All tmax.sample must be contained in data, use drop_trials if necessary.')
    
        length_epoch = self.tmax.sample[0] - self.tmin.sample[0] + 1
        epochs = np.empty((trials.shape[0], data.shape[0], length_epoch))
        for idx, trial in enumerate(trials):
            epochs[idx] = data[:, self.tmin.sample[trial] : self.tmax.sample[trial] + 1]
        return epochs
    



"""
This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""   

from .emgtools import (set_log_file, tkeo, filtfilter, hpfilter, lpfilter, notch_filter, 
                       moving_avg, integrated_profile, 
                       get_onset_ip, 
                       global_var, detector_var, detector_dbl_th, 
                       signal_windows, get_onsets,
                       get_active_sections,
                       get_signal_portions, get_signal_max, somf, get_onset_somf, show_trial)

from .events import (Events, EpochEvents, load_continuous, load_segmented)
                   
from .latency import times, find_times   

from .utils import use_mne, use_txt
from .viz.viz_emg import VizApplication as Viz



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 24 11:40:41 2019

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""   

from .utils import utilsfunc

def samples(tmin, tmax, sf):
    """Compute sample serie.

    Parameters
    ----------
    tmin : float
        Starting time.
    tmax : float
        Ending time.
    sf : int | float
        Sampling frequency.

    Returns
    -------
    sample_serie : 1D array
        Samples array.
    """
    import numpy as np

    if tmin > tmax:
        raise ValueError('tmax must be superior or equal to tmin')

    samplemin = int(np.floor(tmin * sf))
    sample_serie = np.arange(samplemin, int(np.ceil(tmax * sf)) + 1)
    return sample_serie

def times(tmin, tmax, sf):
    """Compute times serie.

    Parameters
    ----------
    tmin : float
        Starting time.
    tmax : float
        Ending time.
    sf : int | float
        Sampling frequency.

    Returns
    -------
    time_serie : 1D array
        Times array in seconds.
    """
    return samples(tmin, tmax, sf) / float(sf)

def find_times(time_value, time_serie):
    """Return index such that timeSerie[index] is the closest value
    of timeSerie to timeValue.

    Parameters
    ----------
    time_value : float
    time_serie : 1D array | list

    Returns
    -------
    index : int

    """
    import numpy as np
    return np.abs(time_serie - time_value).argmin()

def signal_windows(sections, start, end, warn=False):
    """Define mean temporal windows around the given time sections.
    
    The time window for each section starts at the average value between 
    the end of the previous section and the beginning of the current section. 
    The end of the time window corresponds to the start of the next time 
    window, i.e., the average between the end of the current section and 
    the beginning of the next section. Start of first window and end of last 
    windows are defined by the average with start_sample and end_sample.
    If any windows starts before start value, start value is changed to 0. If 
    any window ends after end value, end value is changed to the maximal value
    of sections.
    This function is used for instance to define the different time windows 
    used for EMG onset search.

    Parameters
    ----------
    sections : 2D array
        Each line corresponds to one section. First column gives the section's
        start, second column gives the section's end. 
    start : float
        The beginning of the first time window corresponds to the average
        value between start and the start of first section.
    end : float
        The end of the last time window corresponds to the average
        value between the end of last section and end.
    warn : bool
        If True, a warning is displayed when start value is changed to 0 or
        when end value is changed to maximal value of sections (default False).

    Returns
    -------
    2D array
        Each line corresponds to one time window. First column contains 
        window's start, second columns contains window's end.
    
    """
    import numpy as np

    early_section = sections < start
    late_section = sections > end 

    if any(early_section.flatten()) > 0 : 
        start = 0
        if warn:
            print("Warning: Window starts before start point, new start point is " + str(start))
    if any(late_section.flatten()) > 0 : 
        end  = np.max(sections.flatten())
        if warn:
            print("Warning: Window ends after end point, new end point is" + str(end))
        
    borders = np.transpose((np.hstack((start,sections[:-1,1])),np.hstack((sections[1:,0],end))))
    windows = np.mean((sections,borders),axis=0)
    
    return windows

##############################################################################
# Class Latency ##############################################################
##############################################################################
class Latency():
    """Latency class, contain sample and time latency of events.

    Given latencies are stored in Latency.sample in samples and in Latency.time
    in seconds. Either sample or time can be provided, the other one is
    computed automatically based on sampling frequency (sf). Latency.sample and
    Latency.time are numpy.ndarray, in which one given event is stored at the
    same position (i.e., latency of event 'n' is stored in Latency.sample[n]
    and Latency.sample[n]).

    Parameters
    ----------
    sample : int | list of int | 1D array of int
        Latency in samples.
    time : float | list of float | 1D array of float
        Latency in second.
    sf : float
        Sampling frequency of the corresponding data recording.
    """

    def __init__(self, sample=None, time=None, sf=None):
        import numpy as np

        if sf is None:
            sf = input('Enter sampling frequency: ')
        self.sf = float(sf)

        if (sample is None) & (time is None):
            sample = []
            time = []

        elif (sample is None) & (time is not None):
            time = np.array(utilsfunc.in_list(time))
            if (np.isnan(time)).any():
                print(('Warning: nan in time[{}]!').format(np.where(np.isnan(time))[0]))
            sample = np.array(np.round(time * self.sf), dtype=int)
        elif (sample is not None) & (time is None):
            if (np.isnan(sample)).any():
                print(('Warning: nan in sample[{}]!').format(np.where(np.isnan(sample))[0]))
            sample = np.array(utilsfunc.in_list(sample), dtype=int)
            time = sample / float(self.sf)
        else:
            sample = utilsfunc.in_list(sample)
            time = utilsfunc.in_list(time)

        self.sample = np.array(sample, dtype=int)
        self.time = np.array(time)
        
        if (self.sample.size) != (self.time.size):
            raise TypeError('sample and time must be of same length.')

    def __repr__(self):
        return "class Latency, {} events, sf = {}, sample: {}, time: {}".format(len(self.sample), self.sf, self.sample, self.time)

    def nb_lat(self):
        """Return the number of latencies.
        """
        return len(self.sample)

    def sort_sample(self):
        """Return the indices of sorted samples.
        """
        import numpy as np
        return np.argsort(self.sample)

    def sort_time(self):
        """Return the indices of sorted times.
        """
        import numpy as np
        return np.argsort(self.time)

    def add_lat(self, lat):
        """Add Latency.
        """
        import numpy as np
        if self.sf != lat.sf:
            raise ValueError('Sampling frequencies do not match !')
        self.sample = np.concatenate((self.sample, lat.sample))
        self.time = np.concatenate((self.time, lat.time))

    def del_lat(self, idx_lat):
        """Delete Latency.
        """
        import numpy as np
        self.sample = np.delete(self.sample, idx_lat)
        self.time = np.delete(self.time, idx_lat)

    def get_lat(self, idx_lat):
        """Get the specified latencies.
        """
        return Latency(sample=self.sample[idx_lat], time=self.time[idx_lat], sf=self.sf)

    def copy(self):
        """Return a copy of the current Latency.
        """
        import numpy as np
        copy_sample = np.copy(self.sample)
        copy_time = np.copy(self.time)
        return Latency(sample=copy_sample, time=copy_time, sf=self.sf)

    def find_lat(self, sample=None, time=None):
        """Find the specified latencies and returns the position(s).

        Find the indices of latencies whose sample and/or time are equal to the
        given value(s). If not None, sample and time must be of same length.If 
        both sample and time are filled, find latencies satisfying both 
        requirements: Latency.find_lat(sample = 512, time=0.5) returns latencies 
        indices whose sample == 512 AND time == 0.5. If more than one sample and
        one time are given, search for latencies satisfying each pair of sample/time: 
        Latency.find_lat(sample = [512, 1536], time=[0.5, 1.5]) returns indices of 
        latencies whose (sample == 512 AND time == 0.5) or (sample == 1536 AND
        time == 1.5). Note that using Events.find_events(sample = [512, 1536], time=[0.5, 1.5])
        however returns event latencies whose (sample == 512 OR sample == 1536)
        AND (time == .5 OR time == 1.5).      

        Parameters
        ----------
        sample : int | list of  int | 1D array of int
            Sample value(s) to look for.
        time : float | list of float | | 1D array of float
            Time value(s) to look for.

        Returns
        -------
            Position(s) of the found latencies.
        """

        import numpy as np
        
        if time is None:
            idx_lat = np.flatnonzero(np.in1d(self.sample, sample))
        elif sample is None:
            idx_lat = np.flatnonzero(np.in1d(self.time, time))
        else: 
            sample = utilsfunc.in_list(sample)
            time = utilsfunc.in_list(time)
            if len(sample) != len(time):
                raise ValueError ('Sample and time must of same length.')

            idx_lat = []
            for e in range(len(sample)):
                sample_index = np.flatnonzero(self.sample == sample[e])
                # keeps only those whose time is also equal
                idx_lat.extend([s for s in sample_index if self.time[s] == time[e]])
            idx_lat = np.asarray(idx_lat)
            
        return idx_lat
    



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 13 16:55:42 2016

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""   

def set_log_file(fname, overwrite=False):
    """Set the log to print to a file. 
    Largely inspired by the logging system of mne-python
    ###COPY FROM MNE###
    
    Set the log to print to a file using the 'logging' module. For more information
    about logging module, see: https://docs.python.org/dev/howto/logging.html

    Parameters
    ----------
    fname : str
        Filename of the log to print to.
    overwrite : bool
        Overwrite the log file (if it exists) if True. If False, statements
        will be appended to the log (default). 
        FIXME: define the kwargs used, for example "logging.INFO". 
        Laure: I don't think we should define logging.INFO, it is not a kwarg of our function (it can't be passed as a kwarg to set_log_file). I add 'using logging module'
    """
    import logging
    logger = logging.getLogger('emg_tools')
    logger.setLevel(logging.INFO)
    print(logging.INFO)
    handlers = logger.handlers
    for h in handlers:
        # only remove our handlers (get along nicely with nose)
        if isinstance(h, (logging.FileHandler, logging.StreamHandler)):
            if isinstance(h, logging.FileHandler):
                h.close()
            logger.removeHandler(h)
    mode = 'w' if overwrite else 'a'
    lh = logging.FileHandler(fname, mode=mode)
    logger.addHandler(lh)

def tkeo(array):
    """Return the Teaker-Kayser transformation of the input signal.

    Parameters
    ----------
    array : 1D array
        Signal on which Teager-Kaiser energy operator should be
        computed. If it has more than one dimension, a warning
        is issued and an array of zeros is returned.

    Returns
    -------
    tkeo : array of same dimension as imput array

    References
    ----------
    Li X, Zhou P, Aruin AS (2007) Teager–Kaiser energy operation of surface EMG
    improves muscle activity onset detection. Ann Biomed Eng 35:1532–1538.
    
    Solnik S, Rider P, Steinweg K, DeVita P, HortobaÂgyi T (2010) Teager-Kaiser
    energy operator signal conditioning improves EMG onset detection. Eur J Appl
    Physiol. 110(3):489-98. doi: 10.1007/s00421-010-1521-8

    Tenan MS, Tweedell AJ, Haynes CA (2017) Analysis of statistical and standard
    algorithms for detecting muscle onset with surface electromyography.
    PLoS One. 12(5):e0177312. doi: 10.1371/journal.pone.0177312.
    """
    import numpy as np
    if len(array.shape) > 1:
        print("only 1D vectors are accepted")
        return np.zeros(array.shape)
    else:
        tmp = np.zeros(array.shape)
        tmp[1:-1] = array[1:-1]**2 - (array[:-2] * array[2:])
        tmp[0] = tmp[1]
        tmp[-1] = tmp[-2]
        return tmp

def filtfilter(array, N=3, sf=None, cutoff=None, filter_name=None):
    """
    FIXME: sg.filtfilt is better than sg.lfilter especially for low-pass filtering so 
    I think we should use it instead of hfilter and lfilter, agreed ? -> no actually, hp and low are better because no modifciation at the beginning of the burst
    """
    from numpy import zeros
    import scipy.signal as sg
    if sf is None:
        raise TypeError('Provide suitable value for sampling frequency (sf).')
    if filter_name is None:
        raise TypeError('Provide suitable value for filter_name (\'lowpass\' or \'highass\').')
    Wn = 2. * cutoff / sf
    b, a = sg.butter(N, Wn, filter_name)
    out = zeros(array.shape)
    out = sg.filtfilt(b, a, array)
    return out

def hpfilter(array, N=3, sf=None, cutoff=10):
    """Apply high pass filter to input signal.

    Parameters
    ----------
    array : array_like (1D or 2D)
        Input signal to filter. For 2D arrays, must contain channels on axis 0
        and time samples on axis 1.
    N : int
        The order of the filter (default 3).
    sf : float
        Sampling frequency.
    cutoff : int | float
        Cutoff frequency.

    Returns
    -------
    out : array
        Filtered signal
    """
    from numpy import zeros
    import scipy.signal as sg
    if sf is None:
        raise TypeError('Provide suitable value for sampling frequency (sf).')
    Wn = 2. * cutoff / sf
    b, a = sg.butter(N, Wn, "highpass")
    out = zeros(array.shape)
    out = sg.lfilter(b, a, array)
    return out
    
def lpfilter(array, N=3, sf=None, cutoff=150):  
    """Apply low pass filter to input signal.

    Parameters
    ----------
    array : array_like
        Input signal to filter.For 2D arrays, must contain channels on axis 0
        and time samples on axis 1.
    N : int
        The order of the filter (default 3).
    sf : float
        Sampling frequency.
    cutoff : float
        Cutoff frequency.

    Returns
    -------
    out : array
        Filtered signal
    """
    from numpy import zeros
    import scipy.signal as sg
    Wn = 2. * cutoff / sf
    b, a = sg.butter(N, Wn, "lowpass")
    out = zeros(array.shape)
    out = sg.lfilter(b, a, array)
    return out

def notch_filter(array, N=3, sf=None, freq=50, band=10, ripple=None, filter_type='butter'):
    """Apply notch filter to input signal.

    Parameters
    ----------
    array : array_like
        Input signal to filter. For 2D arrays, must contain channels on axis 0
        and time samples on axis 1.
    N : int
        The order of the filter (default 3). 
    sf : float
        Sampling frequency
    freq : float
        The centerline frequency to be filtered.
    band : float
        The bandwidth around the centerline frequency that you wish to
        filter.
    ripple :
        The maximum passband ripple that is allowed in db
    filter_type : 'butter' | 'bessel' | 'cheby1' | 'cheby2' | 'ellip'
        The filter type to use (default 'butter').

    Returns
    -------
    out : array
        Filtered signal
    """
    import scipy.signal as sg
    nyq = sf / 2.0
    low = freq - band / 2.0
    high = freq + band / 2.0
    low = low / nyq
    high = high / nyq
    b, a = sg.iirfilter(N, [low, high], rp=ripple, btype='bandstop',
                        analog=False, ftype=filter_type)
    out = sg.lfilter(b, a, array)
    return out

def moving_avg(data_trial, window_size):
    """Compute signal moving average with specified window size.

    Parameters
    ----------
    data_trial : 1D array_like
        Input signal.
    window_size : int
        Size of the moving average window in data samples.

    Returns
    -------
    1D signal array after moving average.
    """
    import numpy as np
    data_trial = np.insert(data_trial,0,np.ones(int(window_size - int(window_size)//2))*data_trial[0])
    data_trial = np.append(data_trial,np.ones(int(window_size)//2)*data_trial[-1],0)
    cs = np.cumsum(data_trial)
    return (cs[window_size:] - cs[:-window_size]) / float(window_size)

def integrated_profile(data_trial, times):
    """Return signal integrated profile.

    Compute the signal integrated profile, defined as the difference
    between the empirical cumulative sum of a dataset, and its uniform
    equivalent (straigth line).

    Parameters
    ----------
    data_trial : 1D array
        Input signal
    times : 1D array
        Time instants

    Returns
    -------
    d : 1D array
        Integrated profile
    """
    import numpy as np
    cs = np.abs(data_trial).cumsum()
    a = cs[-1] / (times[-1] - times[0])
    l = a * times
    d = cs - l
    return d

def get_onset_ip(data_trial, times, samples='all', moving_avg_size = 1):
    """Return min and max of signal integrated profile.

    Compute the signal integrated profile and return the position of min
    and max, assumed to correspond to the onset and offset of activity.

    Parameters
    ----------
    data_trial : 1D array
        Input signal.
    times : 1D array
        Time instants.
    samples : 1D array
        Time samples in which min and max are defined. If samples = 'all',
        all samples are used (default is 'all').
    moving_avg_size : int
        Defines the window size (in samples) for the moving average applied 
        on integrated profile before to define min and max. If movingAvg_size 
        is equal to 1, this has no effect (default is 1).

    Returns
    -------
    onset,offset : int
        Indices of data_trial corresponding to EMG onset and offset

    References
    ----------
    
    Liu J, Liu Q (2016) Use of the integrated profile for voluntary muscle
    activity detection using EMG signals with spurious background spikes: A
    study with incomplete spinal cord injury. Biomed Signal Process Control
    24:19-24.
    
    Santello M, McDonagh MJN (1998) The control of timing and amplitude of EMG
    activity in landing movements in humans. Exp Physiol 83:857-874.
    """
    import numpy as np
    if (type(samples) is str) and samples == 'all' : samples = np.arange(len(data_trial))

    ip = integrated_profile(data_trial, times)
    ip = moving_avg(ip[samples],moving_avg_size)
    onset = np.argmin(ip)
    offset = onset + np.argmax(ip[onset:])
    onset = onset + samples[0]
    offset = offset + samples[0]
    return onset, offset

def global_var(data_epochs, times, tmin=None, tmax=None, use_tkeo=True, cor_var=None):
    """Return mean and standard deviation of the input signal between
    tmin and tmax.

    Parameters
    ----------
    data_epochs : 3D array
        First dimension should be trials, second dimension shoud be
        channels, and last dimension is times
    times : 1D array
        Time instants
    tmin : float | None
        Start time for mean/variance computation. If None start at
        first sample (default None).
    tmax : float | None
        End time for mean/variance computation. If None end at time 0
        (default None).
    use_tkeo : bool
        If True, compute mean and variance on the Teaker-Kayser
        transformation of the input signal (default True).
    cor_var : float | None
        If float, correct for outliers. The return mean and variance are
        computed on all sample signals whose abs(z-score) < corVar
        (default None).

    Returns
    -------
    mbsl,stbsl : list
        Mean and standard deviation on each channel.
    """
    import numpy as np
    from .latency import find_times
    if len(data_epochs.shape) != 3:
        print("Data do not have the correct dimensions. Exit !")
        exit()
    mbsl = []
    stbsl = []
    if tmin == None:
        tmin = 0
    if tmax == None:
        tmax = find_times(0, times)
    n_chan = data_epochs.shape[1]
    n_epochs = data_epochs.shape[0]
    for chan in range(n_chan):
        if use_tkeo:
            tkbsl = np.zeros((n_epochs, tmax - tmin))
            for e in range(n_epochs):
                tkbsl[e, :] = tkeo(data_epochs[e, chan, tmin:tmax])
            fbsl = tkbsl.flatten()
        else:
            fbsl = data_epochs[:, chan, tmin:tmax].flatten()
        # correct for outliers
        if cor_var != None:
            mbsl.append(
                fbsl[np.abs((fbsl - fbsl.mean()) / fbsl.std()) < cor_var].mean())
            stbsl.append(
                fbsl[np.abs((fbsl - fbsl.mean()) / fbsl.std()) < cor_var].std())
        else:
            mbsl.append(fbsl.mean())
            stbsl.append(fbsl.std())
    return mbsl, stbsl


def detector_var(data_trial, times, th=3.5, time_limit=.025, 
                 min_samples=3, varying_min=1, mbsl=None, stbsl=None, 
                 baseline='pre_stim', sf=None, use_derivative=False,
                 moving_avg_size=1):
    """Return the indices of time windows containing signal exceeding the 
    specified variance.

    Define the intervals during which one or several portions of input
    signal amplitude exceeds mbsl + (th * stbsl). Intervals are merged when 
    the delay between the interval offset and the next interavl onset is
    inferior to the time limit. Intervals containing less than the defined
    minimal number of samples are excluded.
    For EMG, this gives time windows where EMG bursts are supposed to occur.

    Parameters
    ----------
    data_trial : 1D array
        Input signal. Note that data_trial signal is always rectified before
        detection.
    times : 1D array
        Time instants.
    th : float
        Threshold to multiply by stbsl (default 3.5).
    time_limit : float
        Lowest time delay (in second) between bursts. Bursts intervals are 
        merged When less time separates interval offset from next interval 
        onset (default 0.025s).
    min_samples : float
        Minimal number of data points exceeding mbsl + (th * stbsl) in signal 
        intervals. If less points are present, interval is excluded. 
        This is applied after merging intervals separated by less than the 
        time limit (default 3).
    varying_min : float
        If True, the minimal number of data points exceeding threshold required 
        to define a burst is increased in noisy trials. The noise is estimated 
        by the frequency of "small" bursts (freq_small), i.e., bursts containing
        less than twice the defined minimal number of samples. The minimal 
        number of samples is then increased up to min_samples + 
        (freq_small * varying_min) (Default is 1, set to 0 to have no 
        change of min_samples).
    mbsl : float | None
        If None, use the mean baseline value of the trial (defined by 'baseline'
        parameter). To use global mbsl, use getGlobalVariance function, applied
        on same data type as input signal (e.g., raw EMG, teager-kayser...)
        (default None).
    stbsl : float | None
        If None, use the baseline standard deviation of the trial (defined by 
        'baseline' parameter). To use global stbsl, apply getGlobalVariance on 
        the same data type as input signal (e.g., raw EMG, teager-kayser...)
        (default None).
    baseline : list | array
        Start and end time latencies (in s) to compute baseline mean and standard
        deviation values of the trial (default 'pre_stim' computes mean and 
        standard deviation between first and time 0 samples).
    sf : float
        Sampling frequency.
    use_derivative : bool
        If True signal derivative is used to determine offset (default False).
    moving_avg_size : int
        Defines the window size (in samples) for the moving average applied 
        on signal before detection. If moving_avg_size is equal to 1, this has
        no effect (default is 1).

    Returns
    -------
    2D array
        Each line corresponds to one burst interval. First column contains 
        onset, second column contains offset.
    """
    import numpy as np
    from .latency import find_times
    
    if sf is None:
        raise ValueError('Sampling frequency sf must be provided.')
        
    if baseline == 'pre_stim':
        tbsl = range(find_times(0, times))
    else:
        tbsl = range(find_times(baseline[0], times), find_times(baseline[1], times))
    
    data_trial = moving_avg(data_trial, moving_avg_size)
    data_trial = np.abs(data_trial)

    if mbsl is None: mbsl = data_trial[tbsl].mean()
    if stbsl is None: stbsl = data_trial[tbsl].std()

    above_threshold = data_trial > (mbsl + th * stbsl)
    onsets = np.where((above_threshold[:-1] == False) & (above_threshold[1:] == True))[0] + 1
    offsets = np.where((above_threshold[:-1] == True) & (above_threshold[1:] == False))[0] + 1

    if len(onsets) > 0 or len(offsets) > 0 :
        
        if use_derivative and len(onsets) > 0 and onsets[0] < (data_trial.shape[0] - 1): # we exclude if onset is on last sample
            data_trial = np.abs(np.diff(data_trial))
            
            mbsl = data_trial[tbsl].mean()
            stbsl = data_trial[tbsl].std()

            max_derivative = data_trial[onsets[0]:].argmax() + onsets[0] 
            if data_trial[max_derivative] > (mbsl + th * stbsl):
                offset_derivative = np.where(data_trial[max_derivative:] > (mbsl + th * stbsl))[0]
                if len(offset_derivative) > 0:
                    offsets = offset_derivative[-1:] + max_derivative
                    onsets = onsets[:1]

        if len(offsets) == len(onsets)-1 : #end of signal is above threshold
                offsets = np.append(offsets,len(data_trial)-1) 
        elif len(offsets) == len(onsets)+1 : 
            onsets = np.insert(onsets,0,0) #start of signal is above threshold
    
        intervals = np.where((onsets[1:] - offsets[:-1]) > time_limit * sf)[0]
        onsets = np.hstack((onsets[0],onsets[intervals+1])) ; offsets = np.hstack((offsets[intervals],offsets[-1])) 
 
        samples = np.array([sum(above_threshold[onsets[burst]:offsets[burst]]) for burst in range(len(onsets))])
        
        small_bursts = samples <= min_samples * 2
        per_sec = sum(small_bursts) * sf / float(len(times))
        min_samples = min_samples + (varying_min * per_sec)
        
        onsets = onsets[samples >= min_samples] ; offsets = offsets[samples >= min_samples]
        
    return np.transpose((onsets,offsets))

def detector_dbl_th(data_trial, times, th=3, window_size=.020, 
                    min_above_threshold=.5, min_samples=3, 
                    mbsl=None, stbsl=None, baseline='pre_stim', sf=None):
    """Return the indices of time windows containing signal exceeding the 
    double threshold.    

    Define signal onsets and offsets using a double theshold procedure as 
    proposed by Bonato et al. (1998). Signal onset is detected when a minimum 
    number of points in the specified time window exceeds mbsl + (th * stbsl). 
    Signal offset is detected when the minimum number of points in the 
    time window falls below mbsl + (th * stbsl). Intervals containing less 
    than the defined minimal number of samples are excluded.
    For EMG, this gives time windows where EMG bursts are supposed to occur.

    Parameters
    ----------
    data_trial : 1D array
        Input signal.
    times : 1D array
        Time instants.
    th : float
        Threshold to multiply by stbsl (default 3).
    window_size : float
        Duration of the upfront time window (in second) in which samples 
        exceeding the threshold are detected. 
    min_above_threshold : float
        Minimum proportion of time samples exceeding the theshold in the 
        time window to detect signal (default 50 %).
    min_samples : float
        Minimal number of data points exceeding mbsl + (th * stbsl) in signal 
        intervals. If less points are present, interval is excluded (default 3).
    mbsl : float | None
        If None, use the mean baseline value of the trial (on the time interval 
        defined by 'baseline' parameter). To use global mbsl, use 
        getGlobalVariance function, applied on same data type as input signal 
        (e.g., raw EMG, teager-kayser...) (default None).
    stbsl : float | None
        If None, use the baseline standard deviation of the trial (on the time 
        interval defined by 'baseline' parameter). To use global stbsl, apply 
        getGlobalVariance on the same data type as input signal (e.g., raw EMG,
        teager-kayser...) (default None).
    baseline : list | array
        Start and end time latencies (in s) to compute baseline mean and standard
        deviation values of the trial (default 'pre_stim' computes mean and 
        standard deviation between first time sample and time 0 sample).
    sf : float
        Sampling frequency.

    Returns
    -------
    2D array
        Each line corresponds to one burst interval. First column contains onset, second column contains offset.

    References
    ----------
    Bonato P, D'Alessio T, Knaflitz M (1998) A statistical method for the measurement
    of muscle activation intervals from surface myoelectric signal during gait.
    IEEE Trans Biomed Eng. 45(3):287-99.
    """
    import numpy as np
    from .latency import find_times
    
    if sf is None:
        raise ValueError('Sampling frequency sf must be provided.')


    window_size = int(window_size * sf)
    min_above_threshold = round(min_above_threshold * window_size)
    onsets = np.array([],dtype=int) ; offsets = np.array([],dtype=int)

    if baseline == 'pre_stim':
        tbsl = range(find_times(0, times))
    else:
        tbsl = range(find_times(baseline[0], times), find_times(baseline[1], times))
   
    data_trial = np.abs(data_trial)
    if mbsl is None: mbsl = data_trial[tbsl].mean()
    if stbsl is None: stbsl = data_trial[tbsl].std()

    above_threshold = data_trial > (mbsl + th * stbsl)
    cs = np.cumsum(above_threshold)
    on = np.where((cs[window_size:] - cs[:-window_size]) > min_above_threshold)[0]
    off = np.where((cs[window_size:] - cs[:-window_size]) < min_above_threshold)[0] + window_size
    
    if any(on) : 
        onsets = np.append(onsets,on[0]) 
        if any(off[off > (onsets[-1] + window_size)] < len(data_trial-1)) :
            offsets = np.append(offsets,[off[off > (onsets[-1] + window_size)][0]])
        else: offsets = np.append(offsets,len(data_trial)-1) #end of signal is above threshold
       
        while any(on > offsets[-1]) :
            onsets = np.append(onsets,on[on > offsets[-1]][0])
            if any(off[off > (onsets[-1] + window_size)] < len(data_trial-1)) :
                offsets = np.append(offsets,off[off > (onsets[-1] + window_size)][0] )
            else: offsets = np.append(offsets,len(data_trial)-1) #end of signal is above threshold
    
    samples = np.array([sum(above_threshold[onsets[burst]:offsets[burst]]) for burst in range(len(onsets))])
    onsets = onsets[samples > min_samples] ; offsets = offsets[samples > min_samples]
        
    return np.transpose((onsets,offsets))

def signal_windows(sections, start, end, warn=False):
    """Define mean temporal windows around the given time sections.
    
    The time window for each section starts at the average value between 
    the end of the previous section and the beginning of the current section. 
    The end of the time window corresponds to the start of the next time 
    window, i.e., the average between the end of the current section and 
    the beginning of the next section. Start of first window and end of last 
    windows are defined by the average with start_sample and end_sample.
    If any windows starts before start value, start value is changed to 0. If 
    any window ends after end value, end value is changed to the maximal value
    of sections.
    This function is used for instance to define the different time windows 
    used for EMG onset search.

    Parameters
    ----------
    sections : 2D array
        Each line corresponds to one section. First column gives the section's
        start, second column gives the section's end. 
    start : float
        The beginning of the first time window corresponds to the average
        value between start and the start of first section.
    end : float
        The end of the last time window corresponds to the average
        value between the end of last section and end.
    warn : bool
        If True, a warning is displayed when start value is changed to 0 or
        when end value is changed to maximal value of sections (default False).

    Returns
    -------
    2D array
        Each line corresponds to one time window. First column contains 
        window's start, second columns contains window's end.
    """
    import numpy as np

    early_section = sections < start
    late_section = sections > end 

    if any(early_section.flatten()) > 0 : 
        start = 0
        if warn:
            print("Warning: Window starts before start point, new start point is " + str(start))
    if any(late_section.flatten()) > 0 : 
        end  = np.max(sections.flatten())
        if warn:
            print("Warning: Window ends after end point, new end point is" + str(end))
        
    borders = np.transpose((np.hstack((start,sections[:-1,1])),np.hstack((sections[1:,0],end))))
    windows = np.mean((sections,borders),axis=0)
    
    return windows

def get_onsets(data_trial, times, 
               method='single_threshold', 
               params={},
               baseline='pre_stim',
               use_raw=True, use_tkeo=True, 
               sf=None, ip_search=[-.050,.050], moving_avg_window=.015,
               ):
    """Return burst(s) onset(s) and offset(s).
    
    Identify the time windows in data_trial containing EMG bursts, according to
    the specified method. Then searchs for minimal and maximal values of each 
    window integrated profile.
    Finally, onset is defined as the earliest between the time sample of the 
    minimal value of integrated profile and the EMG burst start as defined by 
    the specified method. Likewise, offset is defined as the latest between the
    sample of maximal value of integrated profile and the end of EMG burst as 
    defined by the specified method. 

    Parameters
    ----------
    data_trial : 1D array
        Input signal.
    times : 1D array
        Time instants.
    method : str
        Defines the function used to detect the presence of EMG burst(s). 
        'single_threshold' uses 'detector_var' function (i.e., detects signal 
         exceeding the specified variance), 'double_theshold' uses 
        'detector_dbl_threshold' function (i.e., signal exceeding the specified 
        variance during a minimal number of time samples) (default is 
        'single_theshold').
    params : dict
        Parameter values used for EMG burst detection using the specified 
        method. For 'single_threshold' method, possible entries (and default 
        values) are: 
        - th_raw : float
            Threshold to multiply by variance, used in 'detector_var' applied
            on raw signal (default 3.5).
        - time_limit_raw : float
            Lowest time delay (in second) between bursts for raw signal. Bursts
            intervals are merged When less time separates interval offset from 
            next interval onset (default is 0.025s).
        - min_samples_raw : float
            Minimal number of data points exceeding threshold to define a burst
            for raw signal. Bursts intervals containing less points exceeding 
            the specified variance are excluded. This is applied after merging 
            intervals separated by less than the time limit (default is 3).
        - varying_min_raw : float
            If True, the minimal number of data points exceeding threshold 
            required to define a burst is increased in noisy trials for raw 
            signal. The noise is estimated by the frequency of "small" bursts 
            (freq_small), i.e., bursts containing less than twice the defined 
            minimal number of samples.The minimal number of samples is then 
            increased up to min_samples + (freq_small * varying_min) (Default 
            is 1, set to 0 to have no change of min_samples).
        - mbsl_raw : float | None
            If None, use the mean baseline value of the raw signal trial (on the 
            time interval defined by 'baseline' parameter). To use global mbsl, 
            use 'global_var' function (default is None).
        - stbsl_raw : float | None
            If None, use the baseline standard deviation of the raw signal trial
            (on the time interval defined by 'baseline' parameter). To use global
            stbsl, apply 'global_var' (default is None).
        - th_tkeo : float
            Threshold to multiply by variance, used in 'detector_var' applied
            on Teager-Kaiser transformation of data_trial signal (default is 8).
        - time_limit_tkeo : float
            Lowest time delay (in second) between bursts for Teager-Kaiser
            transformation of data_trial signal. Bursts intervals are merged 
            when less time separates interval offset from next interval onset
            (default is 0.025s). 
        - min_samples_tkeo : float
            Minimal number of data points exceeding threshold to define a burst
            for Teager-Kaiser transformation of data_trial signal. Bursts
            intervals containing less points exceeding the specified variance 
            are excluded. This is applied after merging intervals separated by 
            less than the time limit (default is 10).
        - varying_min_tkeo : float
            If True, the minimal number of data points exceeding threshold 
            required to define a burst is increased in noisy trials for 
            Teager-Kaiser transformation of data_trial signal. The noise is 
            estimated by the frequency of "small" bursts (freq_small), i.e., 
            bursts containing less than twice the defined minimal number of 
            samples. The minimal number of samples is then increased up to 
            min_samples + (freq_small * varying_min) (Default is 0, meaning no 
            change of min_samples).
        - mbsl_tkeo : float | None
            If None, use the mean baseline value of the Teager-Kaiser 
            transformation of data_trial signal (on the time interval defined by
            'baseline' parameter). To use global mbsl, use 'global_var' function
            on the Teager-Kaiser signal transformation (default is None).
        - stbsl_tkeo : float | None
            If None, use the baseline standard deviation of the Teager-Kaiser
            transformation of data_trial signal (on the time interval defined by
            'baseline' parameter). To use global stbsl, apply 'global_var' on 
            Teager-Kaiser signal transformation (default is None).
        For 'double_threshold method', possible entries (and default values) 
        are: 
        - window_size : float
            Duration of the upfront time window (in second) in which samples 
            exceeding the threshold are detected (default is 0.020s). 
        - min_above_threshold : float
            Minimum proportion of time samples exceeding the theshold in the 
            time window to detect signal (default 0.50).
        - th_raw : float
            Threshold to multiply by stbsl for raw signal (default 3).
        - mbsl_raw : float | None
            If None, use the mean baseline value of raw signal (on the time
            interval defined by 'baseline' parameter). To use global mbsl, use
            'global_var' (default is None).
        - stbsl_raw : float | None
            If None, use the baseline standard deviation of raw signal (on the 
            time interval defined by 'baseline' parameter). To use global stbsl,
            apply 'global_var' (default is None).
        - th_tkeo : float
            Threshold to multiply by stbsl for Teager-Kaiser transformation of
            data_trial signal (default 6).
        - mbsl_tkeo : float | None
            If None, use the mean baseline value of the Teager-Kaiser 
            transformation of data_trial signal (on the time interval defined by
            'baseline' parameter). To use global mbsl, use 'global_var' function 
            on the Teager-Kaiser signal transformation (default is None).
        - stbsl_tkeo : float | None
            If None, use the baseline standard deviation of the Teager-Kaiser
            transformation of data_trial signal (on the time interval defined by
            'baseline' parameter). To use global stbsl, apply 'global_var' on 
            Teager-Kaiser signal transformation (default is None).
        - min_samples : float
            Minimal number of data points exceeding mbsl + (th * stbsl) in 
            signal intervals. If less points are present, interval is excluded 
            (default 3).
    baseline : list | array
        Start and end time latencies (in s) to compute baseline mean and standard
        deviation values of the trial (default 'pre_stim' computes mean and 
        standard deviation between first time sample and time 0 sample).
    use_raw : bool
        If True, apply detection method on raw (i.e., data_trial) signal 
        (default is True).
    use_tkeo : bool
        If True, apply detection method on Teager-Kaiser transformation of
        data_trial signal (default is True).
    sf : float
        Sampling frequency.
    ip_search : 1D array
        Time limits (in second) around EMG burst intervals (as defined by 
        'detector_var') to search for minimal and maximal integrated profile 
        values (Default [-.050 0.050]).
    moving_avg_window : float
        Time window size (in second) of moving average applied to smooth 
        integrated profile before to define minimal and maximal values. To 
        not smooth at all, set moving_avg_window to 1/sf (Default is 0.015).
       
    Returns
    -------
    onsets,offsets : lists 
        Indices of data_trial corresponding to EMG onsets and offsets
    """
    import numpy as np
    from .latency import find_times

    emg_sections = get_active_sections(data_trial, times,\
                                       method=method,\
                                       params=params,\
                                       baseline=baseline,\
                                       use_raw=use_raw, use_tkeo=use_tkeo,\
                                       sf=sf)
                           
    onsets = [] ; offsets = []
    if emg_sections.shape[0] > 0:

        emg_sections = emg_sections[np.argsort(emg_sections[:,0])]

        t0 = find_times(times,0)
        ip_windows = np.array(signal_windows(emg_sections, t0, len(data_trial)),dtype = int)
        bsl_idx = np.hstack((np.arange(0,ip_windows[0,0]),np.arange(ip_windows[-1,1],len(data_trial))))
            
        for b in range(emg_sections.shape[0]):
            
            section_idx = np.hstack((bsl_idx,np.arange(ip_windows[b,0],ip_windows[b,1])))
            section_idx.sort()
            
            sample_search_start = int(emg_sections[b,0] + np.floor(ip_search[0]*sf) - (ip_windows[b,0] - ip_windows[0,0]))
            sample_search_end = int(emg_sections[b,1] + np.ceil(ip_search[1]*sf) - (ip_windows[b,0] - ip_windows[0,0]))
            sample_search = np.arange(np.max((sample_search_start, ip_windows[b,0] - (ip_windows[b,0] - ip_windows[0,0]))), np.min((sample_search_end, ip_windows[b,1] - (ip_windows[b,0] - ip_windows[0,0]))))
            
            on_ip,off_ip = get_onset_ip(data_trial[section_idx], times[:len(section_idx)], samples=sample_search, moving_avg_size=int(moving_avg_window * sf))

            onset = np.min((emg_sections[b,0],on_ip + (ip_windows[b,0] - ip_windows[0,0])))
            offset = np.min((emg_sections[b,1],off_ip + (ip_windows[b,0] - ip_windows[0,0])))
            # FIXME for Laure: decide something between those two and adjust the docstring
            # onset = on_ip + (ip_windows[b,0] - ip_windows[0,0])
            # offset = off_ip + (ip_windows[b,0] - ip_windows[0,0])

            onsets.append(onset)
            offsets.append(offset)
            
    return np.array(onsets,dtype=int), np.array(offsets,dtype=int)

def get_active_sections(data_trial, times,\
                        method='single_threshold',\
                        params={},\
                        baseline='pre_stim',\
                        use_raw=True, use_tkeo=True,\
                        sf=None):
    """ 
    Identify the time windows in data_trial containing EMG bursts, according to
    the specified method. 

    Parameters
    ----------
    data_trial : 1D array
        Input signal.
    times : 1D array
        Time instants.
    method : str
        Defines the function used to detect the presence of EMG burst(s). 
        'single_threshold' uses 'detector_var' function (i.e., detects signal 
         exceeding the specified variance), 'double_theshold' uses 
        'detector_dbl_threshold' function (i.e., signal exceeding the specified 
        variance during a minimal number of time samples) (default is 
        'single_theshold').
    params : dict
        Parameter values used for EMG burst detection using the specified 
        method. For 'single_threshold' method, possible entries (and default 
        values) are: 
        - th_raw : float
            Threshold to multiply by variance, used in 'detector_var' applied
            on raw signal (default 3.5).
        - time_limit_raw : float
            Lowest time delay (in second) between bursts for raw signal. Bursts
            intervals are merged When less time separates interval offset from 
            next interval onset (default is 0.025s).
        - min_samples_raw : float
            Minimal number of data points exceeding threshold to define a burst
            for raw signal. Bursts intervals containing less points exceeding 
            the specified variance are excluded. This is applied after merging 
            intervals separated by less than the time limit (default is 3).
        - varying_min_raw : float
            If True, the minimal number of data points exceeding threshold 
            required to define a burst is increased in noisy trials for raw 
            signal. The noise is estimated by the frequency of "small" bursts 
            (freq_small), i.e., bursts containing less than twice the defined 
            minimal number of samples.The minimal number of samples is then 
            increased up to min_samples + (freq_small * varying_min) (Default 
            is 1, set to 0 to have no change of min_samples).
        - mbsl_raw : float | None
            If None, use the mean baseline value of the raw signal trial (on the 
            time interval defined by 'baseline' parameter). To use global mbsl, 
            use 'global_var' function (default is None).
        - stbsl_raw : float | None
            If None, use the baseline standard deviation of the raw signal trial
            (on the time interval defined by 'baseline' parameter). To use global 
            stbsl, apply 'global_var' (default is None).
        - th_tkeo : float
            Threshold to multiply by variance, used in 'detector_var' applied
            on Teager-Kaiser transformation of data_trial signal (default is 8).
        - time_limit_tkeo : float
            Lowest time delay (in second) between bursts for Teager-Kaiser
            transformation of data_trial signal. Bursts intervals are merged 
            when less time separates interval offset from next interval onset
            (default is 0.025s). 
        - min_samples_tkeo : float
            Minimal number of data points exceeding threshold to define a burst
            for Teager-Kaiser transformation of data_trial signal. Bursts
            intervals containing less points exceeding the specified variance 
            are excluded. This is applied after merging intervals separated by 
            less than the time limit (default is 10).
        - varying_min_tkeo : float
            If True, the minimal number of data points exceeding threshold 
            required to define a burst is increased in noisy trials for 
            Teager-Kaiser transformation of data_trial signal. The noise is 
            estimated by the frequency of "small" bursts (freq_small), i.e., 
            bursts containing less than twice the defined minimal number of 
            samples. The minimal number of samples is then increased up to 
            min_samples + (freq_small * varying_min) (Default is 0, meaning no 
            change of min_samples).
        - mbsl_tkeo : float | None
            If None, use the mean baseline value of the Teager-Kaiser 
            transformation of data_trial signal (on the time interval defined by 
            'baseline' parameter). To use global mbsl, use 'global_var' function 
            on the Teager-Kaiser signal transformation (default is None).
        - stbsl_tkeo : float | None
            If None, use the baseline standard deviation of the Teager-Kaiser
            transformation of data_trial signal (on the time interval defined by
            'baseline' parameter). To use global stbsl, apply 'global_var' on 
            Teager-Kaiser signal transformation (default is None).
        For 'double_threshold method', possible entries (and default values) 
        are: 
        - window_size : float
            Duration of the upfront time window (in second) in which samples 
            exceeding the threshold are detected (default is 0.020s). 
        - min_above_threshold : float
            Minimum proportion of time samples exceeding the theshold in the 
            time window to detect signal (default 0.50).
        - th_raw : float
            Threshold to multiply by stbsl for raw signal (default 3).
        - mbsl_raw : float | None
            If None, use the mean baseline value of raw signal (on the time 
            interval defined by 'baseline' parameter). To use global mbsl, use 
            'global_var' (default is None).
        - stbsl_raw : float | None
            If None, use the baseline standard deviation of raw signal (on the 
            time interval defined by 'baseline' parameter). To use global stbsl, 
            apply 'global_var' (default is None).
        - th_tkeo : float
            Threshold to multiply by stbsl for Teager-Kaiser transformation of
            data_trial signal (default 6).
        - mbsl_tkeo : float | None
            If None, use the mean baseline value of the Teager-Kaiser 
            transformation of data_trial signal (on the time interval defined by 
            'baseline' parameter). To use global mbsl, use 'global_var' function 
            on the Teager-Kaiser signal transformation (default is None).
        - stbsl_tkeo : float | None
            If None, use the baseline standard deviation of the Teager-Kaiser
            transformation of data_trial signal (on the time interval defined by 
            'baseline' parameter). To use global stbsl, apply 'global_var' on 
            Teager-Kaiser signal transformation (default is None).
        - min_samples : float
            Minimal number of data points exceeding mbsl + (th * stbsl) in 
            signal intervals. If less points are present, interval is excluded 
            (default 3).
    baseline : list | array
        Start and end time latencies (in s) to compute baseline mean and standard
        deviation values of the trial (default 'pre_stim' computes mean and 
        standard deviation between first time sample and time 0 sample).
    use_raw : bool
        If True, apply detection method on raw (i.e., data_trial) signal 
        (default is True).
    use_tkeo : bool
        If True, apply detection method on Teager-Kaiser transformation of
        data_trial signal (default is True).
    sf : float
        Sampling frequency.
    """
    import numpy as np
    if method == 'single_threshold':
        if use_raw: 
            # get parameters values for detector_var function
            if 'th_raw' in params.keys(): 
                th_raw = params['th_raw']
            else:
                th_raw = 3.5
            if 'time_limit_raw' in params.keys():
                time_limit_raw = params['time_limit_raw']
            else:
                time_limit_raw = 0.025
            if 'min_samples_raw' in params.keys():
                min_samples_raw = params['min_samples_raw']
            else:
                min_samples_raw = 3
            if 'varying_min_raw' in params.keys():
                varying_min_raw = params['varying_min_raw']
            else:
                varying_min_raw = 1
            if 'mbsl_raw' in params.keys():
                mbsl_raw = params['mbsl_raw']
            else:
                mbsl_raw = None
            if 'stbsl_raw' in params.keys():
                stbsl_raw = params['stbsl_raw']
            else:
                stbsl_raw = None
            
            # apply detector_var (single threshold detector)
            active_sections_raw = detector_var(data_trial, times, 
                                               th=th_raw, time_limit=time_limit_raw, min_samples=min_samples_raw, 
                                               varying_min=varying_min_raw, 
                                               mbsl=mbsl_raw, stbsl=stbsl_raw, baseline=baseline,
                                               sf=sf)
        else: 
            active_sections_raw = np.transpose(([],[]))
        
        if use_tkeo: 
            # get parameters values for detector_var function
            if 'th_tkeo' in params.keys():
                th_tkeo = params['th_tkeo']
            else:
                th_tkeo = 8
            if 'time_limit_tkeo' in params.keys():
                time_limit_tkeo = params['time_limit_tkeo']
            else:
                time_limit_tkeo = 0.025
            if 'min_samples_tkeo' in params.keys():
                min_samples_tkeo = params['min_samples_tkeo']
            else:
                min_samples_tkeo = 10
            if 'varying_min_tkeo' in params.keys():
                varying_min_tkeo = params['varying_min_tkeo']
            else:
                varying_min_tkeo = 0
            if 'mbsl_tkeo' in params.keys():
                mbsl_tkeo = params['mbsl_tkeo']
            else:
                mbsl_tkeo = None
            if 'stbsl_tkeo' in params.keys():
                stbsl_tkeo = params['stbsl_tkeo']
            else:
                stbsl_tkeo = None
                
            # apply detector_var (single threshold detector)
            active_sections_tkeo = detector_var(tkeo(data_trial), times, th=th_tkeo, time_limit=time_limit_tkeo, min_samples=min_samples_tkeo,
                                                varying_min=varying_min_tkeo, 
                                                mbsl=mbsl_tkeo, stbsl=stbsl_tkeo, baseline=baseline,
                                                sf=sf)
        else: 
            active_sections_tkeo = np.transpose(([],[]))
        
    elif method == 'double_threshold':

        # get parameters values for detector_dbl_th function
        if 'window_size' in params.keys():
            window_size = params['window_size']
        else:
            window_size = 0.020
        if 'min_above_threshold' in params.keys():
            min_above_threshold = params['min_above_threshold']
        else:
            min_above_threshold = 0.5
        if 'min_samples' in params.keys():
            min_samples = params['min_samples']
        else:
            min_samples = 3 
        if use_raw: 
            if 'th_raw' in params.keys(): 
                th_raw = params['th_raw']
            else:
                th_raw = 3
            if 'mbsl_raw' in params.keys():
                mbsl_raw = params['mbsl_raw']
            else:
                mbsl_raw = None
            if 'stbsl_raw' in params.keys():
                stbsl_raw = params['stbsl_raw']
            else:
                stbsl_raw = None
                
            # apply detector_dbl_th (double threshold detector)
            active_sections_raw = detector_dbl_th(data_trial, times, th=th_raw, window_size=window_size, 
                                                  min_above_threshold=min_above_threshold, min_samples=min_samples, 
                                                  mbsl=mbsl_raw, stbsl=stbsl_raw, baseline=baseline,
                                                  sf=sf)
        else: 
            active_sections_raw = np.transpose(([],[]))

        if use_tkeo: 
            if 'th_tkeo' in params.keys(): 
                th_tkeo = params['th_tkeo']
            else:
                th_tkeo = 6
            if 'mbsl_tkeo' in params.keys():
                mbsl_tkeo = params['mbsl_tkeo']
            else:
                mbsl_tkeo = None
            if 'stbsl_tkeo' in params.keys():
                stbsl_tkeo = params['stbsl_tkeo']
            else:
                stbsl_tkeo = None
                
            # apply detector_dbl_th (double threshold detector)
            active_sections_tkeo = detector_dbl_th(tkeo(data_trial), times, th=th_tkeo, window_size=window_size, 
                                                   min_above_threshold=min_above_threshold, min_samples=min_samples, 
                                                   mbsl=mbsl_tkeo, stbsl=stbsl_tkeo, baseline=baseline,
                                                   sf=sf)
        else: 
            active_sections_tkeo = np.transpose(([],[]))
        
    for tk in active_sections_tkeo:
        non_overlap = [b for b in range(len(active_sections_raw)) if ((active_sections_raw[b][1] < tk[0]) | (tk[1] < active_sections_raw[b][0]))]
        active_sections_raw = active_sections_raw[non_overlap]

    active_sections = np.vstack((active_sections_raw,active_sections_tkeo))
    
    return np.array(active_sections,dtype=int)

def get_signal_portions(array, start_samples, stop_samples):
    """Get signals between each start_sample and stop_sample.

    Parameters
    ----------
    array : 1D array
        Input signal array.
    start_samples : list | array of int
        Start samples of signal portions to get.
    stop_samples : list | array of int
        Stop (i.e., end) samples of signal portions to get.
    
    Returns
    -------
    list_signals : list
        List of arrays containing signal portions.
    """
    import numpy as np
    start_samples = np.array(start_samples)
    stop_samples = np.array(stop_samples)
    
    if (stop_samples < start_samples).any():
        raise ValueError('All start samples must be inferior to stop samples.')
        
    list_signals = []
    for p in range(len(start_samples)):
        list_signals.append(array[start_samples[p]:stop_samples[p]])
    
    return list_signals
    
def get_signal_max(list_signals):
    """Get signal max of each signal portion.

    Parameters
    ----------
    list_signals : list
        List of arrays containing signal portions (e.g., obtained using
        'get_signal_portions').
    
    Returns
    -------
    val : 1D array
        Array containing maximum value for each array of list_signals.
    lat : 1D array
        Array containing latency of maximum value for each array of list_signals.
    """
    import numpy as np
    
    val = []
    lat = []
    for signal in list_signals:
        val.append(np.max(signal))
        lat.append(np.argmax(signal))
    return np.asarray(val), np.asarray(lat)

def somf(array, W):
    # Laure : do you have a ref? or time to explain ? I have no idea how this
    # is supposed to work + does not run at is in Python 3 because of the division
    
    # FIXME : not thoroughly tested. A warning should be raised!
    #TODO: make (W/2) a variable and use it instead of recomputing all the time
    #TODO2 ? : can't we use an np.array for s instead of a list? If we can 
    # determine its size a priori, it would be faster
    #TODO3: to speed-up, try to use Cython?
    import numpy as np
    s = [(W**2) / 12.]
    array = np.abs(array)
    for ti in np.arange((W / 2), array.shape[0] - (W / 2)):
        num = 0
        denom = 0
        for tn in np.arange(ti - (W / 2), ti + (W / 2)):
            num += (np.abs(array[tn])) * ((tn - ti)**2)
        denom = np.abs(array[ti - (W / 2):ti + (W / 2)]).sum()
        s.append(num / (denom * 1.))
    s = np.array(s)
    return s


def get_onset_somf(data, somf, W=900, L=.5):
    import numpy as np
    muSomf = (W**2) / 12
    th = L * muSomf
    logical = np.zeros(data.shape)

    remainingSignal = somf
    tons = [0]
    toffs = [0]
    while remainingSignal.shape[0] > 0:
        t = np.where(remainingSignal < th)[0]
        if t.shape[0] > 0:
            ton = t.min()
            print('ton = ' + str(ton))
            tons.append(ton + toffs[-1])
            t2 = np.where(remainingSignal[ton:] > th)[0]
            if t2.shape[0] > 0:
                toff = t2.min()
                print('toff = ' + str(toff))
            else:
                toff = data.shape[0]
            toffs.append(toff + tons[-1])
            logical[tons[-1] + W / 2:toffs[-1] + W / 2] = 1
            remainingSignal = somf[toffs[-1]:]
        else:
            remainingSignal = np.array([])
    return tons, toffs


def show_trial(data_trial, times, 
               method='single_threshold',
               params={},\
               baseline='pre_stim',
               use_raw=True, use_tkeo=True, 
               sf=None, ip_search=[-.050,.050], 
               moving_avg_window=.015):
    """Show data trial burst detection and onset/offset detection.
    See 'get_onsets' function for complete description of the parameters.
    """
    # import numpy as np
#    import pylab as plt
    import matplotlib.pyplot as plt
    
    onsets,offsets = get_onsets(data_trial, times, 
                                method=method,
                                params=params,
                                baseline=baseline,
                                use_raw=use_raw, use_tkeo=use_tkeo,  
                                sf=sf, ip_search=ip_search,
                                moving_avg_window=moving_avg_window)

    emg_sections = get_active_sections(data_trial, times,
                                       method=method,
                                       params=params, 
                                       baseline=baseline,
                                       use_raw=use_raw, use_tkeo=use_tkeo,
                                       sf=sf)

    plt.figure()
    plt.plot(times, data_trial, 'k', linewidth = .75)
    for section in emg_sections:
        plt.vlines(times[section[0]], data_trial.min()*1.1, data_trial.max()*1.1, linestyles='dotted')
        plt.vlines(times[section[1]], data_trial.min()*1.1, data_trial.max()*1.1, linestyles='dotted')

    for o in onsets:
        plt.plot(times[o],data_trial[o],'bx', mew=2, markersize=8)

    for o in offsets:
        plt.plot(times[o],data_trial[o],'cx', mew=2,markersize=8)
    
    plt.show()




# -*- coding: utf-8 -*-
"""
Created on Fri Jan 12 11:27:20 2018

@author: Laure Spieser and Boris Burle
Laboratoire de Neurosciences Cognitives
UMR 7291, CNRS, Aix-Marseille Université
3, Place Victor Hugo
13331 Marseille cedex 3

This file is part of Myonset.

Myonset is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Myonset is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Myonset. If not, see <https://www.gnu.org/licenses/>.
"""   

import numpy as np
from .. import emgtools
from . import utilsfunc


def get_ch_idx(raw, ch_names='all'):
    """Get channel indices.
    """
    if ch_names == 'all': 
        ch_idx = list(range(len(raw.ch_names)))
    else:
        ch_idx = [raw.ch_names.index(ch) for ch in ch_names if ch in raw.ch_names]
    return ch_idx

def get_times_idx(raw, tmin=None, tmax=None):
    """Get times indices.
    """
    if tmin == None:
        tmin = 0
    if tmax == None: 
        tmax = raw.times[-1]
    times_idx = np.arange(raw.time_as_index(tmin), raw.time_as_index(tmax))
    return times_idx

def get_data_array(raw, ch_names='all', tmin=None, tmax=None):
    """Get raw data.
    """
    ch_idx = get_ch_idx(raw, ch_names)
    if tmin is None:
        samplemin = 0
    else:
        samplemin = raw.time_as_index(tmin)
    if tmax is not None:
        samplemax = raw.time_as_index(tmax)[0]
    else:
        samplemax = None
    return raw.get_data(ch_idx, start=samplemin, stop=samplemax)

def apply_filter(raw, ch_names='all', N=3, low_cutoff=None, high_cutoff=None):
    """Apply high and low pass filters to input signal.
    
    Parameters:
    -----------
    raw : mne Raw data structure
        Raw mne data containing input signal to filter.
    ch_names : list  | str
        List of channel names on which filter will be applied. If 'all',
        filter is applied on all channels (Default 'all').
    N : int
        The order of the filter (Default 3).
    low_cutoff : float
        Cutoff frequency for high pass filter. If 'None', no high pass filter
        is applied.         
    high_cutoff : float
        Cutoff frequency for low pass filter. If 'None', no low pass filter
        is applied.
        
    Returns:
    --------
    raw_filter : raw mne data structure
        Raw mne data containing filtered signal.
    
    """
    ch_idx = get_ch_idx(raw, ch_names)    
    array = raw._data[ch_idx, :]
        
    # Apply highpass and lowpass filters
    if low_cutoff is not None:
        array = emgtools.hpfilter(array, N=N, sf=raw.info['sfreq'], cutoff=low_cutoff)
        #array = emgtools.filtfilter(array, N=N, sf=raw.info['sfreq'], cutoff=low_cutoff, filter_name='highpass')
    if high_cutoff is not None:
        array = emgtools.lpfilter(array, N=N, sf=raw.info['sfreq'], cutoff=high_cutoff)
        #array = emgtools.filtfilter(array, N=N, sf=raw.info['sfreq'], cutoff=high_cutoff, filter_name='lowpass')
    
    raw_filter = raw.copy()
    raw_filter._data[ch_idx, :] = array
    
    return raw_filter

def bipolar_ref(raw, anode, cathode, new_ch=None, copy=False, ch_info=None):
    """Set bipolar montage.
    """
    from mne import set_bipolar_reference

    cathode = utilsfunc.in_list(cathode)
    if len(cathode) == 1 :
        cathode = cathode * len(utilsfunc.in_list(anode))
    if new_ch is None :
        new_ch = utilsfunc.in_list(anode)
        
    if copy is False:
        set_bipolar_reference(raw,anode=anode,\
                              cathode=cathode,ch_name=new_ch,\
                              copy=False, ch_info=ch_info)
    else:
        return set_bipolar_reference(raw,anode=anode,\
                                     cathode=cathode,ch_name=new_ch,\
                                     copy=True, ch_info=ch_info)

def select_channels(raw, ch_names):
    """Select designated channels.
    """
    return raw.pick_channels(utilsfunc.in_list(ch_names))

def drop_channels(raw, ch_names):
    """Delete designated channels.
    """
    return raw.drop_channels(utilsfunc.in_list(ch_names))

#def get_var(epochs, trials='all', ch_names='all', tmin=None, tmax=None, use_tkeo=True, 
#            cor_var=None):
#    """
#    Return mean and standard deviation of the input signal on specified
#    channels between tmin and tmax.
#
#    Parameters:
#    -----------
#    epochs : mne Epochs data structure
#        Input signal
#    ch_names : list  | str
#        List of channel names to use. If 'all', filter is
#        applied on all channels (Default 'all').
#    tmin : float | None
#        Start time for mean/variance computation. If None start at
#        first sample (default None).
#    tmax : float | None
#        End time for mean/variance computation. If None end at time 0
#        (default None).
#    use_tkeo : bool
#        If True, compute mean and variance on the Teaker-Kayser
#        transformation of the input signal (default True).
#    cor_var : float | None
#        If float, correct for outliers. The return mean and variance are
#        computed on all sample signals whose abs(z-score) < cor_var
#        (default None).
#
#    Returns:
#    --------
#    mbsl,stbsl : list
#        mean and standard deviation on each channel
#
#    """
#        
#    if trials == 'all' : trials = np.arange(epochs._data.shape[0])
#    ch_idx = get_ch_idx(epochs, ch_names)
#    data_epochs = epochs._data[np.ix_(trials,ch_idx)]
#    
#    mbsl,stbsl = emgtools.global_var(data_epochs, epochs.times, tmin=tmin, tmax=tmax, use_tkeo=use_tkeo, cor_var=cor_var)
#    
#    return mbsl, stbsl
    
